
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2008 Yann Glouche and Thomas Genet

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.4. Last modification 28/05/08
*)
(* ..................................................................
    Module.......: Unifystd
    Initial Date : 18 March 2004
    Last Update..: 19 December 2005
    Developer....: Mathieu Turuani
    Maintenance..: Judson Santiago
    Compiler.....: OCAML 3.07 
    Comments.....: Basic term computations, unification, 
                   normalization, purification and substitutions
    Project .....: The CL-Atse Protocol Analyser, AVISPA.
 .................................................................... *)

open Types_unify
open Utils_unify


(* ========================== 
   Basic term Computations 
   ========================== *)


(* The inverse function will receive a term, inverting his type to or
   from type term.PInv *)
let inverse : term -> term = function 
  | PInv t -> t 
  | t -> PInv t
;;


(* Extract the list of variables present in a term. 
   Result added to tas. *)
let rec liste_var (tas : int list) : term list -> int list = function
  | Atm(_)::fin		          -> liste_var tas fin
  | Var(n)::fin		          -> 
      if List.mem n tas then 
	liste_var tas fin 
      else 
	liste_var (n::tas) fin
  | Uplet(l)::fin 
  | Xor(l)::fin                   -> liste_var (liste_var tas l) fin
  | PCrypt(m,k)::fin 
  | SCrypt(m,k)::fin              -> liste_var tas (m::k::fin)
  | PInv(t)::fin                  -> liste_var tas (t::fin)
  | Exp(t,l)::fin                 -> liste_var tas (merge_first (t::fin) l)
  | []                            -> tas
;;


(* Checks if there is any substitution (m,t) that applies to the
   current variable n.  

   Properties of a substitution sigma : 
   - All values are purified.  
   - A value contain only variables of a different theory.  
   - A value is a variable iff this variable has no value. *)
let rec valeur_elem n = function
  | (m,t)::l -> 
      if n=m then 
	t 
      else 
	valeur_elem n l
  | _ -> Var(n)
;;

(* Extracts the substitution "subs" from the structure
   (subs,integer,constraint) and call the function valeur_elem to see
   if there is any substitution in subs that applies to n *)
let valeur (n : int) ((ls,_,_) : t_subst) : term = 
  valeur_elem n ls
;;

(* The number of different atomic types allowed.  The type n 0 is
   reserved for "no type" i.e. the composed type. The type n 1 is
   reserved for "Public/Private Keys" i.e. PInv() *)
(*let nb_AtmTypes = 16;; moved to utils *)

(* Generates a number that will be stored together with the
   Atom/Variable in a tabletype. With this number it is possible to see
   what is the type associated with this Atom/Variable *)
(*let create_number var_id var_type = 
  var_id * nb_AtmTypes + var_type
;; moved to utils *)

(* Creates a variable Var(X) where X is a number identifying the
   variable itself and his type in a table 
let create_var var_id var_type = 
  Var (create_number var_id var_type);;  *)

(* Function to return the sequence number of a variable in the
   tabletype, given the encoded number var_number *)
(*let var_id var_number = 
  var_number / nb_AtmTypes
;; moved to utils *)

(* Get the type of the given variable *)
(*let var_type var_number = 
  var_number mod nb_AtmTypes
;; moved to utils *)

(* Assign a value to a new variable in a substitution, and generate 
   the new variable. The new value has type complex (i.e. no type)  *)
let add_newvalue (ls, FreeVar fv, ctr : t_subst) (t : term) : term * t_subst = 
  let n = create_number fv 0
  in (Var n, ((n,t)::ls, FreeVar (fv+1), ctr))
;;


(* Get the type of the given term *)
let rec term_type (s : t_subst) = function 
  | Atm n -> var_type n
  | Var n -> 
      (match valeur n s with 
	 | Var n -> var_type n 
	 | t -> term_type s t)
  | PInv Atm n -> 
      (match var_type n with 
	 | 1 -> 2
	 | 2 -> 1
	 | _ -> 0)
  | PInv Var n -> 
      (match term_type s (valeur n s) with 
	 | 1 -> 2
	 | 2 -> 1
	 | _ -> 0)
  | _ -> 0
;;

(* Checks if the future value (t) of the variable var_number has the
   same type of the variable. Acceptable values are messages (wich
   have type = 0) or values with the same type of the term t *)
let acceptable_value (sub : t_subst) (var_number : int) (t : term) =
  let vt = var_type var_number in 
    ((vt = 0) || (vt = (term_type sub t)))
;;

(* Merge exponents, simplify when possible. *)
let rec merge_exp tas = function
  | (t,si)::suite ->
      let e = (t, exp_inverse si) in
      let rec g pile = function
	| e2::suite -> if e = e2 then merge suite pile else g (e2::pile) suite
	| [] -> (t,si)::(List.rev tas)
      in merge_exp (g [] tas) suite
  | [] -> tas
;;



(* ===============================
   Applying substitutions to terms  
   =============================== *)


(* Apply a substitution "s" to a term "t", cmp theory *)
let rec clot_norm (s : t_subst) (t : term) : term = 
  match t with
    | Atm(_) as t -> t
    | Var(n)      -> 
	( let t = valeur n s in 
	    match t with 
	      | Var(_) -> t 
	      | _ -> clot_norm s t 
	)
    | PCrypt(m,k) -> PCrypt (clot_norm s m, clot_norm s k)
    | SCrypt(m,k) -> SCrypt (clot_norm s m, clot_norm s k)
    | Uplet(l)	-> 
	Uplet (rev_flat_map 
		 ( function 
		     | Uplet(l) -> (List.rev l) 
		     | t -> [t]
		 ) 
		 [] 
		 (List.rev_map (clot_norm s) l))
    | PInv(t) -> inverse (clot_norm s t)
    | Xor(l)  -> 
	let rec f = function
	  | t::l -> 
	      let l = f l and t = clot_norm s t in 
		if List.mem t l then 
		  l 
		else 
		  (t::l)
	  | _    -> []
	in Xor(f l)
    | Exp(t,l) ->
	let l = List.rev_map (fun (t,si) -> (clot_norm s t, si)) l
	in match clot_norm s t with
	   | Exp(t,ll) -> Exp(t, merge_exp ll l)
	   | t -> Exp(t, List.rev l)
;;

(* Applying a substitution "s" to a term "t", light theory *)
let rec clot_light (s : t_subst) (t : term) : term = 
  match t with
    | Atm(_) as t -> t
    | Var(n) -> 
	( let t = valeur n s in 
	    match t with 
	      | Var(_) as t -> t 
	      | _ -> clot_light s t 
	)
    | PCrypt(m,k) -> PCrypt (clot_light s m, clot_light s k)
    | SCrypt(m,k) -> SCrypt (clot_light s m, clot_light s k)
    | Uplet(l)    -> Uplet (List.map (clot_light s) l)
    | PInv(t)     -> inverse (clot_light s t)
    | Xor(l)      -> 
	let rec f = function
	  | t::l -> 
                let l = f l in
                let t = clot_light s t in 
		(t::l)
	  | _    -> []
	in Xor(f l)
    | Exp(t,l) ->
	let l = List.rev_map (fun (t,si) -> (clot_norm s t, si)) l
	in match clot_light s t with
	   | Exp(t,ll) -> Exp(t, merge_exp ll l)
	   | t -> Exp(t, List.rev l)
;;

(* Chooses which type of clot should be done: normal or light *)
let clot b s t =    (* : bool -> t_subst -> term -> term =  *)
  if !light_unif then 
    clot_light s t
  else 
    match b with 
      | true -> clot_norm s t
      | false -> clot_light s t
;;




(* ======================
   Normalization of Terms
   ====================== *)


(* Comparaison test between NORMALISED and PURIFIED terms v1 and v2.

   Rq : A Xor or Exp is supposed to be simplified and SORTED...                    
   Result : positive if >, 0 if equal, and non positive if <                  
   A variable must always be smaller than everything else.  
   Xor comes next, then PInv, then Atoms, then other things. 
   Rq : Order between Xor and PInv is important for Exp in ctrTest. *)
let comp (s : t_subst) (v1 : term) (v2 : term) : int = 
  let rec f = function
    | (t1::l1,t2::l2) -> 
	( match cmp (t1,t2) with 
	    | 0 -> f (l1,l2) 
	    | x -> x )
    | ([],[])	      ->  0
    | ([],_)	      -> -1
    | (_,[])	      ->  1
  and cmp_exp (a,u) (b,v) = 
    match cmp (a,b) with
      | 0 -> if u=v then 0 else if u=Pos then -1 else 1
      | x -> x
  and g = function
    | (a1::l1, a2::l2) ->
    	( match cmp_exp a1 a2 with
		| 0 -> g (l1,l2)
		| x -> x		)
    | ([],[]) -> 0
    | ([], _) -> -1
    | (_, []) -> 1
  and cmp = function
    | (Var n1, Var n2) -> 
	if n1=n2 then 
	  0 
	else 
	  (match (valeur n1 s, valeur n2 s) with
	     | (Var m1, Var m2) -> m1 - m2  
	     | (Var(_),_) -> -1  
	     | (_,Var(_)) ->  1  
	     | (t1,t2) -> cmp (t1,t2) )
    | (Var n, t) -> 
	( match valeur n s with  
	    | Var(_) -> -1  
	    | tt -> cmp (tt,t) )
    | (t, Var n) -> 
	( match valeur n s with  
	    | Var(_) ->  1  
	    | tt -> cmp (t,tt) )
    | (PInv a, PInv b) -> cmp (a,b)
    | (Xor(l1), Xor(l2))     ->  f (l1,l2)  
    | (Xor(_), _)            -> -1  
    | (_,Xor(_))             ->  1    
    | (PInv _, _) -> -1
    | (_, PInv _) ->  1
    | (Atm n1, Atm n2) -> n1 - n2          
    | (Atm(_), _) -> -1    
    | (_, Atm(_)) ->  1
    | (PCrypt(m1,k1),PCrypt(m2,k2))     -> 
	( match cmp (m1,m2) with 
	    | 0 -> cmp (k1,k2) 
	    | x -> x )
    | (PCrypt(_,_), _) -> -1               
    | (_, PCrypt(_,_)) ->  1
    | (SCrypt(m1,k1),SCrypt(m2,k2))     -> 
	( match cmp (m1,m2) with 
	    | 0 -> cmp (k1,k2) 
	    | x -> x )
    | (SCrypt(_,_), _)       -> -1               
    | (_, SCrypt(_,_))       ->  1
    | (Uplet(l1), Uplet(l2)) ->  f (l1,l2)  
    | (Uplet(_), _)          -> -1
    | (_, Uplet(_))          ->  1
    | (Exp(t1,l1),Exp(t2,l2))           ->
    	( match cmp (t1,t2) with
		| 0 -> g (l1,l2)
		| x -> x			)
  in 
    cmp (v1,v2)
;;


(* Normalise and purify an Uplet (a list of terms) on the standard
   theory by purifying each element. The arguments of the function
   are:

   purifie : t_subst -> term -> (term * t_subst) - function to purify 
   s : t_subst - substitution 
   l : term list - a list of terms representing an Uplet.
   
   The output of this function is (list of terms * substitution), i.e. 
   the new Uplet and substitution. *)
let norm_Uplet_light purifie (s : t_subst) (l : term list) : term list * t_subst =
  let rec norm_Ul = function
    | t::fin -> 
	( let (ll,s2) = norm_Ul fin in 
	    match purifie s2 t with
              | (Var n as tt, ss)           -> 
		  ( match valeur n ss with
		      | Atm(0)
		      | Xor(_) -> (tt::ll, ss)
                      | tt     -> (tt::ll, ss) )    
	      | (Atm(0) as tt, ss)
              | (Xor(_) as tt, ss) -> 
		  let (nv, ns) = add_newvalue ss tt
		  in (nv::ll, ns)
              | (tt, s3)                    -> 
		  (tt::ll, s3)  )
    | []     -> ([], s)
  in 
    norm_Ul l
;;


(* Normalise and purify an Uplet, in the cmp theory.           
   Any variable in an Uplet :                                                  
   - either don't have any value.                                           
   - or has a value in a different theory (like xor).  *)
let norm_Uplet_cmp purifie (sub : t_subst) (lst : term list) : term list * t_subst =
  let rec norm_Uc = function 
    | (Uplet l)::fin -> norm_Uc (l@fin)  (* For <<..>..> = <....> *)
    | tt::fin -> 
	( let (l,ss) = norm_Uc fin in 
	    match purifie ss tt with
              | (Var n as t,s)            -> 
		  ( match valeur n s with
		      | Uplet(ll) -> (ll@l,s)
		      | Atm(0)
		      | Xor(_)    -> (t::l,s)
                      | t         -> (t::l,s) )
              | (Uplet ll,s)              -> (ll@l,s)
	      | (Atm(0) as t, s)
              | (Xor(_) as t, s) -> 
		  let (nv, ns) = add_newvalue s t
		  in (nv::l, ns)
              | (t,s)	                  -> (t::l,s) )
    | [] -> ([],sub)
  in 
    norm_Uc lst
;;


(* Normalise and Purify an Uplet. Choose here the theory you want. *)
let norm_Uplet p s l = 
  if !light_unif then 
    norm_Uplet_light p s l
  else 
    norm_Uplet_cmp p s l
;;


(* Normalise and purify a Xor, in the ACUN theory. The result must be
   sorted.  As usual, any element not a variable (and not a xor) is
   replaced by a variable, but except for atoms.  (to keep the
   substitutions compact)  CHANGE : Atoms too, now.

   Remark : The comparaison/equality test uses the fact that a
   normalised xor is sorted.
   
   Input : a substitution "sub", and a list "lst" of terms
   representing a Xor. *)
let norm_Xor purifie (sub : t_subst) (lst : term list) : term list * t_subst =

  (* Insert an element "t" in a sorted list "l", modulo nilpotence
     (i.e. t::t::l = l) *)
  let rec insere_xor s l t = 
    match l with
      | tt::fin -> ( match comp s t tt with 
                       | 0 -> fin 
                       | x -> 
			   if x<0 then 
			     t::l 
			   else tt::(insere_xor s fin t) )
      | [] -> [t]
  and norm_Xacun = function
    | (Xor l)::fin -> norm_Xacun (l@fin)
    | t::fin -> 
	(let (l,s) = norm_Xacun fin in 
	   match purifie s t with
             | (Var n as t, ss) -> 
		 (match valeur n ss with
                    | Var(_) as tt -> (insere_xor ss l tt, ss)
                    | Xor(ll)      -> (it_list (insere_xor ss) l ll, ss)
                    | Atm(0)       -> (l,ss)
               (*   | Atm(_) as tt -> (insere_xor ss l tt, ss)    Removed for Atom theory separation. *)
                    | _            -> (insere_xor ss l t, ss)  )
	     | (Xor(ll), ss)    -> (it_list (insere_xor ss) l ll, ss)
	     | (Atm 0, ss)      -> (l,ss)
        (*   | (Atm _ as t, ss) -> (insere_xor ss l t, ss)   Removed for Atm theory separation. *)
	     | (t, ss) -> 
		 let (nv,ns) = add_newvalue ss t
		 in (insere_xor ns l nv, ns)     )
    | _ -> ([], sub)
  in 
    norm_Xacun lst
;;



(* Normalise and purify an Exp. The result must be
   sorted.  As usual, any xor is replaced by a variable.

   Remark : The comparaison/equality test uses the fact that a
   normalised exp is sorted.						*)
let rec insere_exp s l t sign = 
  match l with
  | (tt,si2)::fin -> ( match comp s t tt with 
    | 0 -> if sign = si2  then (t,sign)::(tt,si2)::fin  else fin 
    | x -> 
	if x<0 then 
	  (t,sign)::l 
	else (tt,si2)::(insere_exp s fin t sign) )
  | [] -> [t,sign] ;;
let norm_Exp purifie (sub : t_subst) (t : term) (lst : (term * sign) list) : term * t_subst =
  let rec norm_Exp ts = function
    | (t,si)::fin -> 
	(let (l,s) = norm_Exp ts fin in 
	   match purifie s t with
             | (Var n as t, ss) -> 
		 (match valeur n ss with
		    | Xor(_) | Atm(0) -> (insere_exp ss l t si,  ss)
		    | tt              -> (insere_exp ss l tt si, ss)   )
	     | (Xor(_) as t, ss)
	     | (Atm 0 as t, ss) -> 
		 let (nv,ns) = add_newvalue ss t
		 in (insere_exp ns l  nv si, ns)
	     | (tt, ss) -> (insere_exp ss l tt si, ss)	)
    | _ -> ts  in 
  let (lst,sub) = norm_Exp ([],sub) lst  in
   ( match purifie sub t with
      | (Var n as t), sub -> ( match valeur n sub with
				| Exp(t,lst2) -> (* (match t with Var n -> (match valeur n sub with Exp(_,_) -> raise (Failure "No !") | _ -> ()) | _ -> ()) ; *) 
				                 let l,sub = norm_Exp (lst,sub) lst2 
				                 in Exp(t,l), sub  
				| Atm 0 | Xor _ -> Exp(t, lst), sub
				| t -> Exp(t, lst), sub             )
      | Exp(t,lst2), sub -> let l,sub = norm_Exp (lst,sub) lst2 in Exp(t,l), sub
      | (Xor(_) as t), sub
      | (Atm(0) as t), sub -> let (nv,sub) = add_newvalue sub t in Exp(nv, lst), sub
      | t, sub -> Exp(t, lst), sub   )



(* Assign a new variable to a term if it's not standard. (i.e. if a
   Xor). The term is supposed to be Normalised and Purified.  *)
let new_var_if_not_std : term * t_subst -> term * t_subst = function
  | (Atm(0) as e, s)
  | (Xor(_) as e, s) -> add_newvalue s e
  | c -> c
;;


(* Purification and Normalisation of a term with a substitution *)
let rec purify (s : t_subst) : term -> (term * t_subst) = function
  | Atm(_) 
  | Var(_) as e -> (e,s)
  | Uplet(l)	-> ( match norm_Uplet purify s l with
			| ([t],ss) -> (t, ss)
			| (ll, ss) -> (Uplet ll,ss)	)
  | Xor(l)	-> ( match norm_Xor purify s l with
			| ([], ss) -> (Atm 0, ss)
			| ([t],ss) -> (t, ss)
			| (ll, ss) -> (Xor ll, ss)	)
  | Exp(t,l)    -> norm_Exp purify s t l
  | PCrypt(m,k) -> 
      let (mm,s1) = new_var_if_not_std (purify s m)
      in let (kk,s2) = new_var_if_not_std (purify s1 k)
      in (PCrypt(mm,kk), s2)
  | SCrypt(m,k) -> 
      let (mm,s1) = new_var_if_not_std (purify s m)
      in let (kk,s2) = new_var_if_not_std (purify s1 k)
      in (SCrypt(mm,kk), s2)
  | PInv(t)	-> 
      ( match purify s t with
	  | (Var n as t, ss) -> 
	      (match valeur n ss with
                 | PInv(t)  -> (t, ss)
		 | Atm(0)
                 | Xor(_)   -> (PInv t, ss) 
                 | t        -> (PInv t, ss) )
	  | (Atm(0) as t, ss)
	  | (Xor(_) as t, ss) -> 
	      let (nv,ns) = add_newvalue ss t
	      in (PInv nv, ns)
	  | (PInv t, ss)	       -> (t, ss)
	  | (t, ss)		       -> (PInv t, ss)	)
;;


(* Purification with non variable result if possible. 
   (i.e. var => no value) *)
let val_purify (s : t_subst) (t : term ) : term * t_subst = 
  match purify s t with
    | (Var(n), ss) -> (valeur n ss, ss)
    | r -> r
;;


(* Equality test over the NORMALIZED and purified terms 
   Rq : Les Xor(_) sont supposs simplifis et surtout tris. (!) *)
let rec egal (s : t_subst) (v1 : term) (v2 : term) : bool = 
  match (v1,v2) with
    | (Var n1, Var n2) -> 
	if n1=n2 then 
	  true 
	else
	  ( match (valeur n1 s, valeur n2 s) with
	      | (Var m1, Var m2) 	-> m1=m2
	      | (Var(_),_) | (_,Var(_)) -> false
	      | (t1,t2) 		-> egal s t1 t2 )
    | (Var n, t) | (t, Var n) -> 
	(match valeur n s with  
	   | Var(_) -> false  
	   | tt -> egal s t tt)
    | (Uplet(l1), Uplet(l2)) 
    | (Xor(l1), Xor(l2))            -> 
	(try List.for_all2 (egal s) l1 l2 with 
	     _ -> false)
    | (PCrypt(m1,k1),PCrypt(m2,k2)) 
    | (SCrypt(m1,k1),SCrypt(m2,k2)) -> (egal s m1 m2) && (egal s k1 k2)
    | (Atm n1, Atm n2)              -> (n1=n2)
    | (PInv v1, PInv v2)            -> egal s v1 v2
    | (Exp(t1,l1),Exp(t2,l2))       -> 
    	( try 
	    (egal s t1 t2) && 
	    (List.for_all2 (fun (a,u) (b,v) -> (u=v) && (egal s a b)) l1 l2)
	  with 
	    | _ -> false	)
    | _ -> false
;;


(* Compare two terms a and b :
   - Equal      iff for all s compatible with sub, we have s(a) = s(b).
   - NeverEqual if for all s compatible with sub, we have s(a) <> s(b).
   - Uncertain  if we cannot decide this. 
   Rq : Equal need completeness, but NeverEqual may be partial.   *)
let rec ctrTest sub = function
  | (Var n, Var m)          -> 
      if n=m then 
	Equal 
      else 
	( match (valeur n sub, valeur m sub) with 
            | (Var n, Var m)          -> if n=m then Equal else Uncertain
	    | (Var _, _) | (_, Var _) -> Uncertain
            | et                      -> ctrTest sub et  )
  | (Var n, u) | (u, Var n) -> 
      (match valeur n sub with 
	 | Var _ -> Uncertain 
	 | v -> ctrTest sub (u,v) )
  | (Uplet l1, Uplet l2)    ->
      let rec f = function
	| (a::l1, b::l2)  -> 
	    (match ctrTest sub (a,b) with 
	       | Equal -> f (l1,l2)  
	       | res -> res )
	| ([],[])   -> Equal
	| _ -> NeverEqual
      in 
	f (l1,l2)  
  | (Xor l1, Xor l2)    ->	
      let rec f = function
	| ((Var n)::l1, (Var m)::l2) ->
	    if n=m then f (l1,l2) else
	    (match (valeur n sub, valeur m sub) with
	       | (Var n, Var m) 	 -> if n=m then f (l1,l2) else Uncertain
	       | (Var _, _) | (_, Var _) -> Uncertain
	       | (u,v) 			 -> f (u::l1, v::l2)	)
	| ((Var n)::l1, l2) | (l2, (Var n)::l1) -> 
	    (match valeur n sub with
	       | Var _ -> Uncertain
	       | t     -> f (t::l1, l2)				)
	| ((PInv u)::l1, (PInv v)::l2) ->
	    (match (u,v) with
	       | (Var n, Var m) -> if n=m then f (l1,l2) else
		   (match (valeur n sub, valeur m sub) with
		      | (Var n, Var m)		-> if n=m then f (l1,l2) else Uncertain
		      | (Var _, _) | (_, Var _)	-> Uncertain
		      | (u,v)			-> f ((PInv u)::l1, (PInv v)::l2)	)
	       | (Var n, _) -> 
	           (match valeur n sub with
		      | Var _	-> Uncertain
		      | t	-> f ((PInv t)::l1, v::l2)		)
	       | (_, Var n) -> 
	           (match valeur n sub with
		      | Var _	-> Uncertain
		      | t	-> f (u::l1, (PInv t)::l2)		)
	       | _ -> if egal sub u v then f (l1,l2) else
		   (match (u,v) with 
		      | (Xor _, _) | (_, Xor _)	-> Uncertain
		      | (Atm _, _) | (_, Atm _)	-> NeverEqual
		      | _			-> Uncertain	)	(* This can be extended here. *)
	    )
	| ((PInv u)::l1, l2) | (l2, (PInv u)::l1) ->
	    (match u with
	       | Var n -> (match valeur n sub with
			     | Var _ -> Uncertain
			     | t     -> f ((PInv t)::l1, l2)		)
	       | Xor _ -> Uncertain
	       | Atm _ -> NeverEqual
	       | _     -> Uncertain	)				(* This can be extended here. *)
	| ((Atm n)::_, (Atm m)::_) 		-> if n=m then Equal else NeverEqual
	| ((Atm _)::_, _) | (_, (Atm _)::_)   	-> NeverEqual
	| ((PCrypt (m1,k1))::l1, l2) | (l2, (PCrypt (m1,k1))::l1) ->
	    (match (l1,l2) with
	       | (_, (PCrypt (m2,k2))::l2) 	-> if (egal sub m1 m2) && (egal sub k1 k2)
						 then f (l1,l2) 
						 else Uncertain
	       | ((PCrypt (_,_))::_, _)     	-> Uncertain		(* Maybe expendable, but difficult, here. *)
	       | _			     	-> NeverEqual		)
	| ((SCrypt (m1,k1))::l1, l2) | (l2, (SCrypt (m1,k1))::l1) ->
	    (match (l1,l2) with
	       | (_, (SCrypt (m2,k2))::l2)   	-> if (egal sub m1 m2) && (egal sub k1 k2) 
						 then f (l1,l2)
						 else Uncertain
	       | ((SCrypt (_,_))::_, _)      	-> Uncertain		(* Maybe expendable, but difficult, here. *)
	       | _			     	-> NeverEqual		)
	| ((Uplet a)::l1, l2) | (l2, (Uplet a)::l1) ->
	    (match (l1,l2) with
	       | (_, (Uplet b)::l2)   	-> 
		   (try if List.for_all2 (egal sub) a b then f (l1,l2) else Uncertain with
		     | _ -> Uncertain 	)
	       | ((Uplet _)::_, _)      	-> Uncertain		(* Maybe expendable, but difficult, here. *)
	       | _			     	-> NeverEqual		)

	| ((Exp(u,a))::l1, (Exp(v,b))::l2) 		-> 
	    (try if (egal sub u v) && 
	            (List.for_all2 (fun (t1,p1) (t2,p2) -> (p1=p2) && (egal sub t1 t2)) a b) 
	         then f (l1,l2) 
	         else Uncertain 
	     with
	       | _ -> Uncertain 	)
	| ((Exp(_,_))::[], []) | ([], (Exp(_,_))::[])	-> NeverEqual
	| ((Exp(_,_))::_, []) | ([], (Exp(_,_))::_)	-> Uncertain    (* Maybe expendable, but difficult, here. *)
	| ([],[]) 					-> Equal
	| ((Xor _)::_, _) | (_, (Xor _)::_)	-> (print_string "\nWARNING : Bug in ctrTest\n"; Uncertain)
      in								(* Impossible case thanks to normalisation. *)
        f (l1,l2)
	(* Rq : Ok for Xor because free variables appear first, then Xor and PInv, then Atm, etc... *)
	(*      The order in this matching must fit EXACTLY the order in comp ! *)
  | (Exp(t1,l1), Exp(t2,l2)) ->
  	let x = ctrTest sub (t1,t2) in if x<>Equal then Uncertain else
	let rec exists_unif = function    (* Better : maybe_unif *)
	  | (_, []) -> false
	  | ((Var n, si1), l) -> ( match valeur n sub with Var _ -> true | t -> exists_unif ((t,si1),l) )
	  | ((Xor _, _), _) -> true
	  | ((PInv Var n, si1), l) -> ( match valeur n sub with Var _ -> true | t -> exists_unif ((PInv t,si1),l) )
	  | ((PInv Xor _, _), _) -> true
	  | ((PInv (PCrypt(_,_) as t1), si1 as a), (PInv (PCrypt(_,_) as t2), si2)::l)
	  | ((PInv (SCrypt(_,_) as t1), si1 as a), (PInv (SCrypt(_,_) as t2), si2)::l)
	  | ((PInv (Uplet _ as t1), si1 as a), (PInv (Uplet _ as t2), si2)::l)
	  | ((PInv (Exp(_,_) as t1), si1 as a), (PInv (Exp(_,_) as t2), si2)::l)
	  | ((PCrypt(_,_) as t1, si1 as a), (PCrypt(_,_) as t2, si2)::l)
	  | ((SCrypt(_,_) as t1, si1 as a), (SCrypt(_,_) as t2, si2)::l)
	  | ((Uplet _ as t1, si1 as a), (Uplet _ as t2, si2)::l)
	  | ((Exp(_,_) as t1, si1 as a), (Exp(_,_) as t2, si2)::l)
	      -> if (si1<>si2) && ((ctrTest sub (t1,t2)) <> NeverEqual) then true
	         else exists_unif (a,l) 
	  | _ -> false  in 
	let rec g = function
	  | Var n -> (match valeur n sub with Var _ -> 0 | t -> g t)
	  | Xor _ -> 0
	  | PInv t -> (g t) / 10
	  | Atm _       -> 10
	  | PCrypt(_,_) -> 20
	  | SCrypt(_,_) -> 30
	  | Uplet(_)    -> 40
	  | Exp(_,_)    -> 50   in
	let rec f = function
	  | ([],[]) -> Equal
	  | (_, []) | ([], _) -> NeverEqual
	  | ((t1,si1 as a)::l1, (t2,si2 as b)::l2) ->
	      match (ctrTest sub (t1,t2), si1=si2) with
	      | Equal,true -> f (l1,l2)
	      | Uncertain,true -> Uncertain
	      | _ -> let g1 = g t1 and g2 = g t2 in
		     if ((exists_unif (a,l1)) && (g2 >= g1))
		     || ((exists_unif (b,l2)) && (g1 >= g2))
		     then Uncertain
	             else NeverEqual
	in f (l1,l2)
  | (SCrypt(m1,k1),SCrypt(m2,k2)) 
  | (PCrypt(m1,k1),PCrypt(m2,k2)) -> 
      ( match ctrTest sub (m1,m2) with 
          | Equal      -> ctrTest sub (k1,k2)
	  | NeverEqual -> NeverEqual
	  | Uncertain  -> 
	      match ctrTest sub (k1,k2) with 
		| NeverEqual -> NeverEqual 
		| _ -> Uncertain )
  | (Atm n,  Atm m ) -> if n=m then Equal else NeverEqual
  | (PInv u, PInv v) -> ctrTest sub (u,v)
  | ((Xor _ as u), t) | (t, (Xor _ as u)) -> ctrTest sub (u, Xor [t])
  | (PInv Var n, t) | (t, PInv Var n) -> (match valeur n sub with
  					    | Var _ | Xor _ -> Uncertain
					    | PInv tt -> ctrTest sub (tt,t) 
                                            | _ -> NeverEqual )
  | (PInv Xor _, _) | (_, PInv Xor _) -> Uncertain

  | (PInv (PInv u), v) | (u, PInv (PInv v)) -> ctrTest sub (u,v)
(*
  | (Uplet _, PCrypt(_,_)) | (Uplet _, SCrypt(_,_)) | (Uplet _, Atm _) | (Uplet _, Exp(_,_)) | (Uplet _, PInv(Atm _))
  | (Uplet _, PInv(Uplet _)) | (Uplet _, PInv(PCrypt(_,_))) | (Uplet _, PInv(SCrypt(_,_))) | (Uplet _, PInv(Exp(_,_)))
  | (PCrypt(_,_) , Uplet _) | (PCrypt(_,_) , SCrypt(_,_)) | (PCrypt(_,_) , Atm _) | (PCrypt(_,_) , Exp(_,_)) | (PCrypt(_,_) , PInv(Atm _))
  | (PCrypt(_,_) , PInv(Uplet _)) | (PCrypt(_,_) , PInv(PCrypt(_,_))) | (PCrypt(_,_) , PInv(SCrypt(_,_))) | (PCrypt(_,_) , PInv(Exp(_,_)))
  | (SCrypt(_,_) , PCrypt(_,_)) | (SCrypt(_,_) , Uplet _) | (SCrypt(_,_) , Atm _) | (SCrypt(_,_) , Exp(_,_)) | (SCrypt(_,_) , PInv(Atm _))
  | (SCrypt(_,_) , PInv(Uplet _)) | (SCrypt(_,_) , PInv(PCrypt(_,_))) | (SCrypt(_,_) , PInv(SCrypt(_,_))) | (SCrypt(_,_) , PInv(Exp(_,_)))
  | (Atm _ , PCrypt(_,_)) | (Atm _ , SCrypt(_,_)) | (Atm _ , Uplet _) | (Atm _ , Exp(_,_)) | (Atm _ , PInv(Atm _))
  | (Atm _ , PInv(Uplet _)) | (Atm _ , PInv(PCrypt(_,_))) | (Atm _ , PInv(SCrypt(_,_))) | (Atm _ , PInv(Exp(_,_)))
  | (Exp(_,_) , PCrypt(_,_)) | (Exp(_,_) , SCrypt(_,_)) | (Exp(_,_) , Atm _) | (Exp(_,_) , Uplet _) | (Exp(_,_) , PInv(Atm _))
  | (Exp(_,_) , PInv(Uplet _)) | (Exp(_,_) , PInv(PCrypt(_,_))) | (Exp(_,_) , PInv(SCrypt(_,_))) | (Exp(_,_) , PInv(Exp(_,_)))
  | (PInv(Atm _) , PCrypt(_,_)) | (PInv(Atm _) , SCrypt(_,_)) | (PInv(Atm _) , Atm _) | (PInv(Atm _) , Exp(_,_)) | (PInv(Atm _) , Uplet _)
  | (PInv(Uplet _) , Uplet _) | (PInv(Uplet _) , PCrypt(_,_)) | (PInv(Uplet _) , SCrypt(_,_)) | (PInv(Uplet _) , Atm _) | (PInv(Uplet _) , Exp(_,_))
  | (PInv(PCrypt(_,_)) , Uplet _) | (PInv(PCrypt(_,_)) , PCrypt(_,_)) | (PInv(PCrypt(_,_)) , SCrypt(_,_)) | (PInv(PCrypt(_,_)) , Atm _) | (PInv(PCrypt(_,_)) , Exp(_,_))
  | (PInv(SCrypt(_,_)) , Uplet _) | (PInv(SCrypt(_,_)) , PCrypt(_,_)) | (PInv(SCrypt(_,_)) , SCrypt(_,_)) | (PInv(SCrypt(_,_)) , Atm _) | (PInv(SCrypt(_,_)) , Exp(_,_))
  | (PInv(Exp(_,_)) , Uplet _) | (PInv(Exp(_,_)) , PCrypt(_,_)) | (PInv(Exp(_,_)) , SCrypt(_,_)) | (PInv(Exp(_,_)) , Atm _) | (PInv(Exp(_,_)) , Exp(_,_)) 
      -> NeverEqual    (* This big matching must have the same effect as  _ -> NeverEqual  *)
*)
  | _ -> NeverEqual  
and ctrTest_exp sub (a,u) (b,v) = 
  if u<>v then NeverEqual else ctrTest sub (a,b)
;;


(* Add a list of negative constraints to a substitution *)
let rec add_ctr (s : t_subst)  : t_unequal -> t_subst = function
  | [] -> s
  | (a,b)::fin -> 
      (* Purify first term *)
      let (a,ss) = purify s a in 
	
      (* Purify second term *)
      let (b,(ls,fv,ctr as sub)) = purify ss b
      in
	(* Test the negative constraint *)
	match ctrTest sub (a,b) with 
	  | Equal -> raise No_Solution
          | NeverEqual -> add_ctr sub fin 
	  | Uncertain -> add_ctr (ls,fv,(a,b)::ctr) fin
;;


(* Test if (Var m) appear in the values of n::l, w.r.t the subst. s *)
let rec recursif s m = function
  | n::l -> 
      ( match valeur n s with
          | Var n -> 
	      if n=m then 
		true 
	      else 
		recursif s m l
          | t     -> recursif s m ((liste_var [] [t])@l) 
      )
  | []   -> false
;;


(* Test if (Var m) appears in a term, w.r.t the subst. s. 
   Rq : "seen" is a list of variables to prevent testing twice the same.
        "r" is true iff we found m under a xor already.
   It returns the following :
     false : (Var m) does not appear in t,
       i.e. in the "add" function it is possible to add (m,t) to the subst.
     true : (Var m) appears in t but only under a Xor,
       i.e. in "add" function this can vanish and we need to use unif_combi.
     raise No_Solution : (Var m) appears in the values of t but NOT below a Xor,
       i.e. it cannot vanish and therefore it is not possible to add (m, t) to the subst.    *)
let rec contains_var (s : t_subst) (m : int) (r : bool) (seen : int list) : term list -> bool = function
  | (Var n)::fin -> if List.memq n seen then contains_var s m r seen fin else
            ( match (valeur n s) with
                | Var nn -> if nn=m then raise No_Solution else contains_var s m r (n::nn::seen) fin
                | t      -> contains_var s m r (n::seen) (t::fin)  )
  | (Atm _)::fin   -> contains_var s m r seen fin
  | (Xor l)::fin   -> contains_var s m (r || (recursif s m (liste_var [] l))) seen fin
  | (Uplet l)::fin -> contains_var s m r seen (merge fin l)
  | (PCrypt (a,b))::fin 
  | (SCrypt (a,b))::fin  -> contains_var s m r seen (a::b::fin)
  | (Exp(t,l))::fin -> contains_var s m r seen (merge_first (t::fin) l)
  | (PInv t)::fin  -> contains_var s m r seen (t::fin)
  | [] -> r;;


(* Exception raised by add_elem when it need external help to unify a var with a term. *)
exception NeedXorUnification of int * term * t_subst ;;


(* Add a value to a free variable in a substitution.
   Input :  to_purify : Do the term need to be purified ?                
           var_number : variable identifier.                             
                  ter : the term to add on nn.                           
                subst : the substitution to be modified.                 
   Output : The new substitution. *)
let rec add_elem (to_purify : bool) (var_number : int) (ter : term) (subst : t_subst) : 
  t_subst =
  let g_Base ee = (ee,false) 
  in

  let g_Uplet var_nn lt ee = 
    match ee with
      | (n,Uplet l) -> ((n, Uplet (replace var_nn lt l)),false)
      | _	    -> (ee,false)  
  in

  let g_PInv nn t ee = 
    match ee with (* Here n:PubKey => m:PubKey => t:PubKey *)
      | (n,PInv (Var m)) -> if m <> nn then (ee,false) else ((n,t), true)
      | _		 -> (ee,false)  
  in

  let valeur2 n ls1 ls2 = 
    match valeur_elem n ls1 with
      | Var m as t -> if m=n then valeur_elem n ls2 else t
      | t          -> t  
  in

  (* lst : modified elements in the substitution to be evaluated. 
     tas : part of the substitution already computed. *)
  let rec f (nn,tt,sub,g as elem) lst tas = function
    | (n,Var m as ee)::fin -> 
	if m=nn then 
	  let ee = (n,tt) in 
	    f elem (ee::lst) (ee::tas) fin 
	else 
	  f elem lst (ee::tas) fin
    | (n,Xor l)::fin -> 
        ( match norm_Xor (fun s t -> (t,s)) (sub,FreeVar 0,[]) l with
          (* Rq : The new subst. is equal to sub since l is already purified. *)
            | ([Var m],_) -> 
		let t = valeur2 m tas fin in 
		  f elem ((n,t)::lst) ((n,t)::tas) fin
	    | ([t],_)     -> f elem ((n,t)::lst) ((n,t)::tas) fin
	    | ([],_)      -> f elem ((n,Atm 0)::lst) ((n,Atm 0)::tas) fin
	    | (l,_)       -> f elem lst ((n,Xor l)::tas) fin               
	)
    | (n,Exp(t,l))::fin -> 
        ( match norm_Exp (fun s t -> (t,s)) (sub,FreeVar 0,[]) t l with
          (* Rq : The new subst. is equal to sub since t and l are already purified. *)
	    | (Exp(_,_) as t,_) -> f elem lst ((n,t)::tas) fin
	    | (t, _) -> f elem ((n,t)::lst) ((n,t)::tas) fin
	)
    | ee::fin -> 
	let (e2,b) = g ee in 
	  ( match e2 with
              | (n, Var m) -> 
		  let t = valeur2 m tas fin in 
		    f elem (if b then (n,t)::lst else lst) ((n,t)::tas) fin
	      | _          -> f elem (if b then e2::lst else lst) (e2::tas) fin   
	  )
    | _ -> 
	(match lst with
           | (nn,(Uplet(lt) as tt))::fin -> 
	       f (nn,tt,tas, g_Uplet (Var nn) lt ) fin [] tas
	   | (nn,(PInv(t) as tt))::fin   -> 
	       f (nn,tt,tas, g_PInv  nn t        ) fin [] tas
	   | (nn,tt)::fin	         -> 
	       f (nn,tt,tas, g_Base              ) fin [] tas
	   | []			         -> 
	       tas				       
	)   
  in

  let (tt,(ls,fv,ctr as s)) = 
    if to_purify then 
      val_purify subst ter 
    else
      match ter with 
	| Var(n) -> (valeur n subst, subst) 
	| _      -> (ter,subst)
  in 
    if tt = (Var var_number) then 
      s 
    else
      if (not (acceptable_value s var_number tt)) then 
	(* But, a value failing to this test has a second chance : if
	   t=PInv(y) with y of type 0 (i.e. any mesage), then maybe
	   adding y = PInv(x) instead would work. HUM : I do not see
	   why there is not the same case for Xor: if t=Xor(y,z,...)
	   with y and z of type 0, we could try y=Xor(x,...)  or
	   z=Xor(x,...). And even both if one fail with No_Solution.
	   This has to be checked.  *)
	match tt with
	  | PInv Var m -> 
	      (match valeur m s with 
	         | Var m -> 
		     if var_type m = 0 then 
		       add_elem false m (PInv (Var var_number)) s 
		     else 
		       raise No_Solution 
		 | _ -> raise No_Solution 
	      )

          (* Added for bug correction. Check for Pb. *)
	  | Var m -> 
	      (match valeur m s with 
	        | Var m -> 
		    if var_type m = 0 then 
		      add_elem false m (Var var_number) s
		    else
		      raise No_Solution
		| _ -> raise No_Solution
	      )

          | Xor _      -> raise (NeedXorUnification (var_number, tt, s))
	  | _          -> raise No_Solution  
      else
	(* Second, the main condition is that x \not\in s(t). This is
	   tested by "recursif", or recursive in english, i.e. true iff
	   "x" appear in the argument, w.r.t "s".  *)
	(* if (recursif s var_number (liste_var [] [tt])) then  *)
	(*   raise No_Solution  *)
	
	(* New version : the main condition is that x \not\in s(t). This
	   is tested by contains_var, i.e. raise No_Solution iff "x" appear in 
	   the argument, w.r.t "s", outside a Xor; "true" if we need to
	   call the unification algorithm modulo Xor because x appear under
	   a Xor only; and "false" iff x do not appear in s(t).    *)
	if contains_var s var_number false [] [tt] then
	  raise (NeedXorUnification (var_number, tt, s))
	else 

	  (* When all these conditions are ok, we can start the
	     simplification > of "s". This is the function "f" that do this. This
	     function pass over > the substitution, using it's argument
	     "(nn,tt,sub,g as elem)" to > normalize the parts that need to be. Rq :
	     The Xor is normalized > anyway, due to it's properties. Maybe we can
	     be more efficient here.  > Arguments : "f (nn,tt,sub,g as elem) lst
	     tas subst".  > - "subst" is the part of the substitution that still
	     need to be checked.  > - "tas" is the part of the substitution that
	     has been checked. We move > elements from subst to tas.  > - "lst" is
	     the list of changes in the substitution that require to repass > over
	     all the substitution again. It's a list of (variable, value). We add >
	     an element (x,t) in lst when the new value of t can produce simpli- >
	     fications in other variables values, i.e. when the new value of x
	     belong > to a different purification class than the old one. On the
	     other hand, > when we know that a value change connot have
	     side-effects, we do not > add it to lst.  > > Initially, lst is
	     [var_number,tt], i.e. the first change in "s". Also, "tas" > contain
	     "s" and the new value x=t. In order to run, "f" must know > the
	     "(nn,tt,sub,g as elem)" adapted to x=t. Since "f" can compute > this
	     itself, the initial value of "subst" is an empty set to force "f" > to
	     compute extract the first element of lst (i.e. x=t) and compute > the
	     corresponding element "(nn,tt,sub,g as elem)". No need to > write this
	     code twice. :) > So, the initial value (0,Atm 0,[],fun x -> (x,false))
	     is just fake.  >
	  *)
	  add_ctr ( (f (0,Atm 0,[],fun x -> (x,false)) 
		       [var_number,tt] 
		       ((var_number,tt)::ls) 
		       []),
		    fv,
		    []
		  ) ctr
;;


(* Add a value to a variable for each substitution in a list.  If the
   variable already has a value, use the unification function provided.
   Result added to tas.  *)
let add unif unif_combi (to_purify : bool) (n : int) (t : term) 
  (tas : t_subst list) (subs : t_subst list) = 
  let rec add_sub tas = function
    | s::suite -> 
	( match valeur n s with
	      (* No substitution for the variable n *) 
	    | Var(nn) -> 
		(try add_sub ((add_elem to_purify nn t s)::tas) suite with 
		   | No_Solution -> add_sub tas suite 
		   | NeedXorUnification(nn,tt,ss) -> 
		       add_sub (merge tas (unif_combi false (Var nn) tt [ss])) suite
		 )
	    | tt      -> add_sub (merge tas (unif to_purify tt t [s])) suite )
    | _        -> tas
  in 
    add_sub tas subs
;;



(* ======================
   Unification algorithms
   ====================== *)


(* Unification algorithm, standard theory, no associativity for pair *)
(*
let rec unif_light (u : term) (v : term) (subs : t_subst list) : t_subst list =
  match (u,v) with
    | (Var(n),_)                        -> add (fun tp -> unif_light) true n v [] subs
    | (_,Var(n))                        -> add (fun tp -> unif_light) true n u [] subs
    | (Atm(n1),Atm(n2))                 -> if n1=n2 then subs else []
    | (Uplet(l1),Uplet(l2))             -> it_list2 unif_light [] subs (l1,l2)
    | (PCrypt(m1,k1),PCrypt(m2,k2))     -> unif_light m1 m2 (unif_light k1 k2 subs)
    | (SCrypt(m1,k1),SCrypt(m2,k2))     -> unif_light m1 m2 (unif_light k1 k2 subs)
    | (PInv(t1),PInv(t2))               -> unif_light t1 t2 subs
    | (PInv(Var(n)),t) 
    | (t,PInv(Var(n))) -> unif_light (Var n) (PInv t) subs *)(* t <> PInv(_) *)(*
    | (PInv(PInv u),v) 
    | (u,PInv(PInv v)) -> unif_light u v subs
    | (Xor(l1),Xor(l2))                 -> it_list2 unif_light [] subs (l1,l2)  *)(* No prop. for xor here.*)(*
    | _                                 -> []
;;
*)

(* Unification algorithm, in the cmp (complete) theory. 
   The terms are purified depending on to_purify.  *)
(*
let rec unif_cmp (to_purify : bool) (u : term) (v : term) (subs : t_subst list) : 
  t_subst list =
  match (u,v) with
    | (Var(n),_)            -> add unif_cmp to_purify n v [] subs
    | (_,Var(n))            -> add unif_cmp to_purify n u [] subs
    | (Atm(n1),Atm(n2))     -> if n1=n2 then subs else []
    | (Uplet(lu),Uplet(lv)) -> 
	let rec f z1 z2 ss = 
	  if ss=[] then 
	    ss 
	  else 
	    match (z1,z2) with
              | ([],[])               -> ss
              | (Var(_)::_,Var(_)::_) -> (g ss z1 [] z2)@(g ss z2 [List.hd z1] (List.tl z1))
              | (Var(_)::_,_)         -> g ss z1 [] z2
              | (_,Var(_)::_)         -> g ss z2 [] z1
              | (t1::zz1,t2::zz2)     -> f zz1 zz2 (unif_cmp false t1 t2 ss)
              | _                     -> []
	and 
	  g su d1 d2 = function
            | [] -> []
            | t::d3 -> 
		(f (List.tl d1) d3 (unif_cmp false (List.hd d1) 
				      (if d2=[] then 
					 t 
				       else 
					 Uplet (List.rev (t::d2))) su))
		@ (g su d1 (t::d2) d3)
	in 
	  if to_purify then 
	    let h s = 
	      *)(* Purify the term u keeping the substitutions used for
		 purification in s1 *)(*
	      let (uu,s1) = 
		purify s u in 
		*)(* Purify the term v keeping the substitutions used for
		   purification in s2 *)(*
	      let (vv,s2) = purify s1 v in 
		match (uu,vv) with
		  | (Uplet l1, Uplet l2)	-> f l1 l2 [s2]
		  | (t1, t2)		-> unif_cmp false t1 t2 [s2]
	    in rev_flat_map h [] subs
	  else 
	    f lu lv subs
    | (PCrypt(m1,k1),PCrypt(m2,k2))     -> 
	unif_cmp to_purify m1 m2 (unif_cmp to_purify k1 k2 subs)
    | (SCrypt(m1,k1),SCrypt(m2,k2))     -> 
	unif_cmp to_purify m1 m2 (unif_cmp to_purify k1 k2 subs)
    | (Uplet([t1]),_)                   -> 
	unif_cmp to_purify t1 v subs
    | (_,Uplet([t2]))                   -> 
	unif_cmp to_purify u t2 subs
    | (PInv(t1),PInv(t2))               -> 
	unif_cmp to_purify t1 t2 subs
    | (PInv(t1),t2) | (t2,PInv(t1))     -> 
	( match t1 with
	    | PInv(tt)    -> unif_cmp to_purify tt t2 subs
	    | Var(_)      -> unif_cmp true t1 (PInv t2) subs
	    | Uplet([tt]) -> unif_cmp to_purify (inverse tt) t2 subs
	    | Xor([tt])   -> unif_cmp true (inverse tt) t2 subs
	    | _           -> [] )
    | (Xor(lu),Xor(lv))                 -> 
	if to_purify then 
	  rev_flat_map (fun s -> 
			  let (uu,s1) = 
			    purify s u in 
			  let (vv,s2) = purify s1 v 
                          in unif_cmp false uu vv [s2]  ) [] subs
	else it_list2 (unif_cmp false) [] subs (lu,lv)
    | (Xor([t1]),_) -> unif_cmp to_purify t1 v subs
    | (_,Xor([t2])) -> unif_cmp to_purify u t2 subs
    | _             -> []
;;
*)

(* Unification choice *)
(*
let unif p u v s = 
  if !light_unif then 
    unif_light u v s
  else 
    unif_cmp p u v s
;;
*)
