
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2008 Yann Glouche and Thomas Genet

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.4. Last modification 28/05/08
*)
(**
unifytest.ml

Link with ATSE Unification Module
@author Yann GLOUCHE
@version 1.0
@since 12/12/2005
*)


(* ==========================
   Link with ATSE Unification
   ========================== *)

open Unify;;
open Types_unify;;
open Utils_unify;;
open Unifyxor;;
open Utils_simulator;;
open Automate;;
open Variables;;

open Debugprinting;;



light_unif := false;;

(**
create an ATSE number
@param n is an integer
@return int
*)
let cn (n : int) : int = create_number n 0;;
(**
create an ATSE variable with identifier "n", with no type
@param n is an identifier
@return an ATSE variable (Types_unify.term)
*)
let x (n : int) : Types_unify.term = Var(create_number n 0);;
(**
create an ATSE value (atom) with identifier "n", with no type
@param n is an identifier
@return an ATSE atom (Types_unify.term)
*)
let a (n : int) : Types_unify.term = Atm(create_number n 0);;
(**
create an ATSE variable with identifier "n", with the type "t"
@param n is an identifier
@param t is the type identifier
@return an ATSE variable (Types_unify.term)
*)
let xt (n : int) (t : int) : Types_unify.term = Var(create_number n t);;
(**
create an ATSE value (atom) with identifier "n", with th type "t"
@param n is an identifier
@param t is the type identifier
@return an ATSE atom (Types_unify.term)
*)
let at (n : int) (t : int) : Types_unify.term = Atm(create_number n t);;


(**
for test the unification
*)
let test_unif 	(unifier : bool -> Types_unify.term -> Types_unify.term -> Types_unify.t_subst list -> Types_unify.t_subst list) 
		(tab : (int * string) list) (str : string) (u : Types_unify.term) (v : Types_unify.term) (sub : Types_unify.t_subst) :  Types_unify.t_subst list =
  let () = print_string ("\n\n\n===== "^str^
             " ==========================================\n"); flush stdout
  in
  let t_start = Sys.time () in
  let subs = unifier true u v [sub] in
  let t_stop = Sys.time () in
  let t_diff = t_stop -. t_start in
  let size = string_of_int (List.length subs) in
  let () = print_string ("number of unifiers: "^size^"\n") in
  let () = print_string ("time needed: "^(if t_diff=0. then "<0.01"
                                          else string_of_float t_diff)^"s\n") in
  (* debug ---> *)
  let () = print_string (
        "std unif alg: "^(if !light_unif then "unif_light" else "unif_cmp")^
        "\n\n==== given problem ===="^
        "\nu = "^(str_term_dbg tab u)^
        "\nv = "^(str_term_dbg tab v)^
        "\n---- sub ----\n"^(str_sub_dbg tab sub)^
        "\n===== solution(s) =========================================\n"^
        (str_subs_dbg tab subs size)^
        "\n===== solution(s) clot ====================================\n"^
        (str_subs_clot_dbg tab subs size u v)^"\n")
  in
  (* <--- debug *)
  let () = print_string (
        "===== /"^str^
        " =========================================\n\n")
  in
  flush stdout; subs
;;

(**
for test the unification
*)
let test (tab : (int * string) list) (str : string) (u : Types_unify.term) (v : Types_unify.term) (sub : Types_unify.t_subst) : Types_unify.t_subst list =
	test_unif unif tab str u v sub;;

(**
@param s is an ATSE term
@return write the ATSE term s
*)
let display (s : Types_unify.term) : string = 
	let subs = unif true (x 0) s [([], FreeVar 100, [])] in
	let size = string_of_int (List.length subs) in
	let l_subs =  lst_subs_dbg subs size in
	let rec aux l =
		match l with [] -> ""
			|(x::ls) -> 	if((List.length l)=1) 
					then (str_term_dbg [] (snd (car x)))
					else aux ls
	in aux l_subs
;;

(** contains the transcrption of an IF term to an ATSE term *)
let automtat_to_unify_format = (Hashtbl.create 600 : (string * string, string * Types_unify.term) Hashtbl.t) ;;
(** contains the transcrption of an ATSE term to an IF term *)
let unify_format_to_automat = (Hashtbl.create 600 : (string, string * string) Hashtbl.t);;
(** contains the translation of an IF type to an ATSE type, the keys of the hash table is an IF type and the value is an ATSE type *)
let automate_types_to_unify_types = (Hashtbl.create 600 : (string, int) Hashtbl.t);;

(**
  used for generate an ATSE term with an IF value or an IF variable
*)
let id_var_atom = (ref 0 : int Pervasives.ref);;
(**
  used for generate an ATSE term with an IF type (nat, text, nonce,...)
*)
let id_type = (ref 1 : int Pervasives.ref);;

(**
compute the pattern of the message of the receiver in th ATSE format
@param name is an agent
@param mess is a message
@param untyped_variable is a boolean, if untyped_variable=true then the type of the variables are not take into account, else the type of the variables is take into account
@return an Term of the ATSE format (Types_unify.term)
*)
let to_unify_format_reciever (name : string) (mess : string) (untyped_variable : bool) : (string * string) * Types_unify.term =
	let give_unify_type type_automat = 
		let dec_type_automat = decompose type_automat in (* we consider: all the sets have the same type  *)
		let type_automat0 = if((dec_type_automat)<>[] && (((car dec_type_automat)="set") || ((car dec_type_automat)="list"))) then (car dec_type_automat) else type_automat in
		if(type_automat0="message" || ((dec_type_automat<>[]) && (type_automat0<>"set") && (type_automat0<>"list"))) then 0
		else  if(Hashtbl.mem automate_types_to_unify_types type_automat0) then (Hashtbl.find automate_types_to_unify_types type_automat0)
		else (id_type := 1+ (!id_type) ; Hashtbl.add automate_types_to_unify_types type_automat0 (!id_type) ; (!id_type) )
	in
	let add mess1 mess2 =
		let display_mess2 = display mess2 in
		if(((snd mess1)<>"") && not(Hashtbl.mem  automtat_to_unify_format mess1 ) )
		then(
		Hashtbl.add  automtat_to_unify_format mess1 (display_mess2,mess2);
		(*prerr_string ((fst mess1)^","^(snd mess1)^" -> "^(str_term_dbg [] mess2)^"\n");*)
		);
		if(((snd mess1)<>"") && not(Hashtbl.mem  unify_format_to_automat display_mess2) )
		then(
		Hashtbl.add  unify_format_to_automat display_mess2 mess1;
		(*prerr_string ((fst mess1)^","^(snd mess1)^" -> "^(str_term_dbg [] mess2)^"\n");*)
		);
		(mess1,mess2)
	in
	let add_var e =
		let h = (Hashtbl.find_all  automtat_to_unify_format (name,e) ) in
		if(h<>[]) then ((name,e),(snd (car h)))
		else (  id_var_atom := 1+(!id_var_atom) ;
			add (name,e) (x (!id_var_atom))
		)
	in
	let add_atom e =
		let h = (Hashtbl.find_all  automtat_to_unify_format ("_",e) ) in
		if(h<>[]) then(("_",e),(snd (car h)))
		else (  id_var_atom := 1+(!id_var_atom) ;
			add ("_",e) (a (!id_var_atom))
		)
	in
	let add_var_typed e t =
		let h = (Hashtbl.find_all  automtat_to_unify_format (name,e) ) in
		if(h<>[]) then((name,e),(snd (car h)))
		else (  id_var_atom := 1+(!id_var_atom) ;
			add (name,e) (xt (!id_var_atom) t)
		)
	in
	let add_atom_typed e t =
		let h = (Hashtbl.find_all  automtat_to_unify_format ("_",e) ) in
		if(h<>[]) then(("_",e),(snd (car h)))
		else (  id_var_atom := 1+(!id_var_atom) ;
			add ("_",e) (at (!id_var_atom) t)
		)
	in
	let rec aux mess =
		let hmess = Hashtbl.find_all automtat_to_unify_format (name,mess) in
		if(hmess<>[]) then ((name,mess), (snd (car hmess)))
		else(

			if ( (is_a_value mess)) then (add_atom mess)
			else if (untyped_variable &&(is_a_test_var mess)) then (
				let value = (get_value_of_var name mess) in
				if((decompose value)<>[]) then (
					let ((_,m01),m02) = aux value in 
					add (name,m01) m02
					)
				else add_atom value
					)
			else if (untyped_variable &&(is_a_var mess)) then (add_var mess)
			else if (is_a_value mess) then (add_atom_typed (get_name_of_var mess) (give_unify_type (get_type mess)))
			else if (is_a_test_var mess) then (
				let value = (get_value_of_var name mess) in
				if((decompose value)<>[]) then (
					let ((_,m01),m02) = aux value in 
					add (name,m01) m02
					)
				else add_atom_typed value (give_unify_type (get_type (get_name_of_var mess)))
					)
			else if (is_a_var mess) then (add_var_typed (get_name_of_var mess) (give_unify_type (get_type (get_name_of_var mess))))
			else
			let l_mess = decompose mess in
			if(l_mess=[]) then (add_var mess)
			else
			(if((car l_mess)="pair") then (
							let l = List.map aux (cdr l_mess) in
							let l_fst = List.map (fun e -> snd (fst e)) l in
							let l_snd = List.map snd l in
							add (name, ("pair("^(list2string "," l_fst)^")"))
							(Uplet l_snd)
				)
			else if((car l_mess)="scrypt") then (
							let ((_,m01),m02) = aux (cadr l_mess) in
							let ((_,m11),m12) = aux (caddr l_mess) in
							add (name, ("scrypt("^m01^","^m11^")"))
							(SCrypt (m12, m02))
				)
			else if((car l_mess)="crypt") then (
							let ((_,m01),m02) = aux (cadr l_mess) in
							let ((_,m11),m12) = aux (caddr l_mess) in
							add (name, ("crypt("^m01^","^m11^")"))
							(PCrypt (m12, m02))
				)
			else if((car l_mess)="inv") then (
							let ((_,m01),m02) = aux (cadr l_mess) in
							add (name, ("inv("^m01^")"))
							(PInv (m02))
				)
			else if((car l_mess)="xor") then (	
							let l = List.map aux (cdr l_mess) in
							let l_fst = List.map (fun e -> snd (fst e)) l in
							let l_snd = List.map snd l in
							add (name, ("xor("^(list2string "," l_fst)^")"))
							(Xor (l_snd))
				)
			else if((car l_mess)="exp") then (
							let l = List.map aux (cdr l_mess) in
							let l_fst = List.map (fun e -> snd (fst e)) l in
							let l_snd = List.map snd l in
							add (name, ("exp("^(list2string "," l_fst)^")"))
							(Exp ((car l_snd), (List.map (fun ((_,a),e) -> if(a.[0]='-') then (e, Neg) else  (e, Pos) )(cdr l))))
				)
			else if((car l_mess)="apply") then (
							let l = List.map aux (cdr l_mess) in
							let l_fst = List.map (fun e -> snd (fst e)) l in
							let l_snd = List.map snd l in
							add (name, ("apply("^(list2string "," l_fst)^")"))
							(SCrypt ((Uplet (cdr l_snd)), (car l_snd)))
				)
			else (
							let l = List.map aux l_mess in
							let l_fst = List.map (fun e -> snd (fst e)) l in
							let l_snd = List.map snd l in
							add (name, ((car l_fst)^"("^(list2string "," (cdr l_fst))^")"))
							(PCrypt ((Uplet (cdr l_snd)), (car l_snd)))
				)
			) 
		);
	in
	aux mess
;;


(**
compute the pattern of the message of the sender in th ATSE format
@param name is an agent
@param mess is a message
@param untyped_variable is a boolean, if untyped_variable=true then the type of the variables are not take into account, else the type of the variables is take into account
@return an Term of the ATSE format (Types_unify.term)
*)
let to_unify_format_sender (name : string) (mess : string) (untyped_variable : bool) : (string * string) * Types_unify.term =
	let give_unify_type type_automat = 
		let dec_type_automat = decompose type_automat in (* we consider: all the sets have the same type  *)
		let type_automat0 = if((dec_type_automat)<>[] && (((car dec_type_automat)="set") || ((car dec_type_automat)="list"))) then (car dec_type_automat) else type_automat in
		if(type_automat0="message" || ((dec_type_automat<>[]) && (type_automat0<>"set") && (type_automat0<>"list"))) then 0
		else  if(Hashtbl.mem automate_types_to_unify_types type_automat0) then (Hashtbl.find automate_types_to_unify_types type_automat0)
		else (id_type := 1+ (!id_type) ; Hashtbl.add automate_types_to_unify_types type_automat0 (!id_type) ; (!id_type) )
	in
	let add mess1 mess2 =
		let display_mess2 = display mess2 in
		if(((snd mess1)<>"") && not(Hashtbl.mem  automtat_to_unify_format mess1 ) )
		then(
		Hashtbl.add  automtat_to_unify_format mess1 (display_mess2,mess2);
		(*prerr_string ((fst mess1)^","^(snd mess1)^" -> "^(str_term_dbg [] mess2)^"\n");*)
		);
		if(((snd mess1)<>"") && not(Hashtbl.mem  unify_format_to_automat display_mess2 ) )
		then(
		Hashtbl.add  unify_format_to_automat display_mess2 mess1;
		(*prerr_string ((fst mess1)^","^(snd mess1)^" -> "^display_mess2^"\n");*)
		);
		(mess1,mess2)
	in
	let add_var e =
		let h = (Hashtbl.find_all  automtat_to_unify_format (name,e) ) in
		if(h<>[]) then ((name,e),(snd (car h)))
		else (  id_var_atom := 1+(!id_var_atom) ;
			add (name,e) (x (!id_var_atom))
		)
	in
	let add_atom e =
		let h = (Hashtbl.find_all  automtat_to_unify_format ("_",e) ) in
		if(h<>[]) then(("_",e),(snd (car h)))
		else (  id_var_atom := 1+(!id_var_atom) ;
			add ("_",e) (a (!id_var_atom))
		)
	in
	let add_var_typed e t =
		let h = (Hashtbl.find_all  automtat_to_unify_format (name,e) ) in
		if(h<>[]) then((name,e),(snd (car h)))
		else (  id_var_atom := 1+(!id_var_atom) ;
			add (name,e) (xt (!id_var_atom) t)
		)
	in
	let add_atom_typed e t =
		let h = (Hashtbl.find_all  automtat_to_unify_format ("_",e) ) in
		if(h<>[]) then(("_",e),(snd (car h)))
		else (  id_var_atom := 1+(!id_var_atom) ;
			add ("_",e) (at (!id_var_atom) t)
		)
	in
	let rec aux mess =
		let hmess = Hashtbl.find_all automtat_to_unify_format (name,mess) in
		if(hmess<>[] && not(is_a_var mess) && ((is_a_value mess) || (is_a_refresh_var mess))) then ((name,mess), (snd (car hmess)))
		else(

			if (untyped_variable && (is_a_value mess)) then (add_atom mess)
			else if (untyped_variable &&(is_a_var mess)) then (
				(*if (is_a_refresh_var mess) then	add_atom (give_next_refresh_value name mess (!tag_number))
				else *)
				let value = (get_value_of_var name mess) in
				if((decompose value)<>[]) then (
					let ((_,m01),m02) = aux value in 
					add (name,m01) m02
					)
				else add_atom value
				)
			else if (is_a_value mess) then (add_atom_typed mess (give_unify_type (get_type mess)))
			else if (is_a_var mess) then (
				let t = (get_type (get_name_of_var mess)) in 
				(*if (is_a_refresh_var mess) then add_var_typed mess (give_unify_type t)
				else*)
				let value = (get_value_of_var name mess) in
				if((decompose value)<>[]) then (
					let ((_,m01),m02) = aux value in 
					add (name,m01) m02
					)
				else add_atom_typed value (give_unify_type t)
				)
			else
			let l_mess = decompose mess in
			if(l_mess=[]) then (add_var mess)
			else
			(if((car l_mess)="pair") then (
							let l = List.map aux (cdr l_mess) in
							let l_fst = List.map (fun e -> snd (fst e)) l in
							let l_snd = List.map snd l in
							add (name, ("pair("^(list2string "," l_fst)^")"))
							(Uplet l_snd)
				)
			else if((car l_mess)="scrypt") then (
							let ((_,m01),m02) = aux (cadr l_mess) in
							let ((_,m11),m12) = aux (caddr l_mess) in
							add (name, ("scrypt("^m01^","^m11^")"))
							(SCrypt (m12, m02))
				)
			else if((car l_mess)="crypt") then (
							let ((_,m01),m02) = aux (cadr l_mess) in
							let ((_,m11),m12) = aux (caddr l_mess) in
							add (name, ("crypt("^m01^","^m11^")"))
							(PCrypt (m12, m02))
				)
			else if((car l_mess)="inv") then (
							let ((_,m01),m02) = aux (cadr l_mess) in
							add (name, ("inv("^m01^")"))
							(PInv (m02))
				)
			else if((car l_mess)="xor") then (	
							let l = List.map aux (cdr l_mess) in
							let l_fst = List.map (fun e -> snd (fst e)) l in
							let l_snd = List.map snd l in
							add (name, ("xor("^(list2string "," l_fst)^")"))
							(Xor (l_snd))
				)
			else if((car l_mess)="exp") then (
							let l = List.map aux (cdr l_mess) in
							let l_fst = List.map (fun e -> snd (fst e)) l in
							let l_snd = List.map snd l in
							add (name, ("exp("^(list2string "," l_fst)^")"))
							(Exp ((car l_snd), (List.map (fun ((_,a),e) -> if(a.[0]='-') then (e, Neg) else  (e, Pos) )(cdr l))))
				)
			else if((car l_mess)="apply") then (
							let l = List.map aux (cdr l_mess) in
							let l_fst = List.map (fun e -> snd (fst e)) l in
							let l_snd = List.map snd l in
							add (name, ("apply("^(list2string "," l_fst)^")"))
							(SCrypt ((Uplet (cdr l_snd)), (car l_snd)))
				)
			else (
							let l = List.map aux l_mess in
							let l_fst = List.map (fun e -> snd (fst e)) l in
							let l_snd = List.map snd l in
							add (name, ((car l_fst)^"("^(list2string "," (cdr l_fst))^")"))
							(PCrypt ((Uplet (cdr l_snd)), (car l_snd)))
				)
			) 
		);
	in
	aux mess
;;

(**
compute an unification of the ATSE term "u" of the agent "n1" with the ATSE term "v" of the agent "n2". "u" can be a message send by "n1" or a state of "n1", and "v" can be a message receive by "n2" or a state of "n2"
@param n1 is an agent
@param u is an ATSE term (Types_unify.term)
@param n2 is an agent
@param u is an ATSE term (Types_unify.term)
@return an unification. An unification is a liste of 4-uplet (a0, v0, n, a) where the variable "v" of the agent "n" take the value of the variable "v0" of the agent "a0" or the value of the value "v0". 
*)
let unification (u : Types_unify.term) (v : Types_unify.term) 
		(n1 : string) (n2 : string) (sub : Types_unify.t_subst) : (string * string * string * string) list * bool = 
	(*let subs = (unif_combi unif unif_xor) true u v [sub] in*)

	let rec sub_max l res = 
		match l with [] -> res
			|(x::ls) -> if((List.length x)<(List.length res)) then sub_max ls x
					else sub_max ls res
	in
	let subs = unif true v u [sub] in
	let size = string_of_int (List.length subs) in
	let l_subs =  lst_subs_dbg subs size in
	(*prerr_string ("\n---\n"^(str_subs_dbg [] subs size)^"\n---\n");*)
	((if(l_subs<>[])
	then(	
		let f t =
		let res = ref ("","") in
		Hashtbl.iter
		(fun a (v,term_atse) ->
			if( (!res)=("","") && term_atse<>t && (unif true t term_atse [sub])<>[] )
			then res:=a
		)
		automtat_to_unify_format;
		!res
		in
		remove ("","","","") (
			List.map
			(fun (var,value) ->
				let hvar = Hashtbl.find_all unify_format_to_automat (str_term_dbg [] var) in
				let(name01,var_automat1) = if(hvar<>[]) then (car hvar) else (f var) in
				let hvalue = Hashtbl.find_all unify_format_to_automat (str_term_dbg [] value) in
				let(name02,var_automat2) = if(hvalue<>[]) then (car hvalue) else (f value) in
					
				if((n2=name01 ) && not(is_a_test_var var_automat1) && (is_a_var var_automat1)) then(
					let name2 = if (name02="_") then name01 else name02 in
					let value_automat = if((is_a_var var_automat2) && not(is_a_refresh_var var_automat2)) then (get_value_of_var name01 var_automat2) else var_automat2 in
					(name2,value_automat,name01,var_automat1)
				)
				else 	if((n2=name02) && not(is_a_test_var var_automat2) && (is_a_var var_automat2)) then(
					let name1 = if (name01="_") then name02 else name01 in
					let value_automat = if((is_a_var var_automat1) && not(is_a_refresh_var var_automat1)) then (get_value_of_var name02 var_automat1) else var_automat1 in
					(name1,value_automat,name02,var_automat2)
				) 
				else ("","","","")
 			)
			(if (l_subs<>[]) then (sub_max (cdr l_subs) (car l_subs)) else [])
		)
 	)
	else []),(subs<>[]))
;;

(**
compute an unification of the message "mesage1" send by the agent "name1" with the message "message2" receive by the agent "name2". If the unification doesn't exist, the function return [].
@param name1 is an agent
@param message1 is a message send by "name1"
@param name2 is an agent
@param message2 is a message receive by "name2"
@return an unification. An unification is a liste of 4-uplet (a0, v0, n, a) where the variable "v" of the agent "n" take the value of the variable "v0" of the agent "a0" or the value of the value "v0".
*)
let unify (name1 : string) (message1 : string) (name2 : string) (message2 : string) (untyped_variable : bool) : (string * string * string * string) list =
	if(message1="" || message2="") then []
	else(   Hashtbl.clear automtat_to_unify_format;
		Hashtbl.clear unify_format_to_automat;
(*	prerr_string (sender_format name1 message1); prerr_string "\n";*)
		let ((_,mu),u) = (to_unify_format_sender name1 message1 untyped_variable) in
		let ((_,mv),v) = (to_unify_format_reciever name2 message2 untyped_variable) in
	(*	if(containStr "ka" message1) then(prerr_string (message1^" = "^mu^" =\n");*)
	(*	prerr_string (message2^" = "^mv^" =\n");); *)
	(*	prerr_string (message1^" = "^mu^" =\n"^(str_term_dbg [] u)^"\n");
		prerr_string (message2^" = "^mv^" =\n"^(str_term_dbg [] v)^"\n -- \n");
		test [] "test" u v ([] ,FreeVar 100, []);*)
		if(u=v) then [("","","","")]
		else if(mu="" || mv="") then []
		else(
			let (res,t) = unification u v name1 name2 ([], FreeVar 100, []) in
		(*	if(res<>[]) then *) (* ( prerr_string (name1^" "^mu^" "^name2^" "^mv^" "^(string_of_bool (res<>[]))^"\n")) ;*)
		(*	List.iter (fun (a,b,c,d) -> prerr_string(a^" "^b^" "^c^" "^d)) res; prerr_string ("\n\n");*)
			 (*test [] "test" u v ([] ,FreeVar 100, []);*)
		(*	List.iter (fun (a,b,c,d) -> prerr_string (a^" - "^b^" - "^c^" - "^d^" -\n") ) res;*)

		(*prerr_string (message1^" = "^mu^" =\n"^(str_term_dbg [] u)^"\n");
		prerr_string (message2^" = "^mv^" =\n"^(str_term_dbg [] v)^"\n -- \n");
		List.iter (fun (a,b,c,d) -> prerr_string(a^" "^b^" "^c^" "^d^"\n")) res;*)

		(*	if(not(t)) then (
				light_unif:=false;
				let (res2,t2) = unification u v name1 name2 ([], FreeVar 100, []) in
				light_unif:=true;
				if(res2=[] && t2) then [("","","","")] else res2;
			)
			else *)
			if(res=[] && t) then [("","","","")] else res;
		)

	)
;;

(***** ***** ***** ***** *****)


(** Match messages *)

(** 
contains all the unifications computed in the current state of the protocol (or the current state the automate)
- the key is a 7-uplet (name of the first agent, state of the first agent, message send by the first agent, name of the second agent, state of the second agent, message of the second agent, the value of variable !untyped_variables)
- the value is an unification
*)
 let match_message_story = (ref (Hashtbl.create 600) : (string * string * string * string * string * string * bool, (string * string * string * string) list) Hashtbl.t Pervasives.ref);;
 (** it is the previous value of !match_message_story. Thus, it is possible to come back to the previous value of !match_message_story *)
 let latest_message_story = (ref (Hashtbl.create 600) : (string * string * string * string * string * string * bool, (string * string * string * string) list) Hashtbl.t Pervasives.ref);;
 (**
 save !match_message_story in !latest_message_story, and clear !match_message_story. Thus, it is possible to come back to the previous value of !match_message_story
 @return unit
 *)
 let clear_message_story (() : unit) : unit = 
	let tmp = (!match_message_story) in
	match_message_story:=(!latest_message_story);
	latest_message_story:=tmp;
	Hashtbl.clear (!match_message_story)
;;
(**
the hash table !match_messate_story came back at her previous value
@return unit
*)
 let previous_message_story (() : unit) =
	let tmp = (!match_message_story) in
	match_message_story:=(!latest_message_story);
	latest_message_story:=tmp
 ;;
 (** 
 save an unification in the hash table !match_message_story
 @param name is the name of an agent
 @param state is a state of the agent "name"
 @param m1 is a message of the agent "name"
 @param name2 is the name of an agent
 @param state2 is a state of the agent "name2"
 @param m2 is a message of the agent "name2"
 @return unit
 *)
 let add_in_match_message_story (name : string) (state : string) (m1 : string) (name2 : string) (state2 : string) (m2 : string) (untyped_variable : bool) (boolean : (string * string * string * string) list) : unit = 
	Hashtbl.add (!match_message_story) (name,state,m1,name2,state2,m2,untyped_variable) boolean;;
(**
 return the result of an unification, if it is previously compute and save in the hash table !match_message_story
 @param name is the name of an agent
 @param state is a state of the agent "name"
 @param m1 is a message of the agent "name"
 @param name2 is the name of an agent
 @param state2 is a state of the agent "name2"
 @param m2 is a message of the agent "name2"
 @return a pair (int, unif). If int=-1 then the unification is not computed yet, else if int=0 the unification is impossible, else if int=1 the unification exist and unif it is this unification.
*)
 let get_match_message_story (name : string) (state : string) (m1 : string) (name2 : string) (state2 : string) (m2 : string) (untyped_variable : bool) : (int * (string * string * string * string) list) =
	match (Hashtbl.find_all (!match_message_story) (name,state,m1,name2,state2,m2,untyped_variable))
	with 	[] -> (-1,[])
		|([]::_) -> (0,[])
	     	|(x::_) -> (1,x)
	;;
(** unify a message "m1" send by an agent  "name" in a state "sender_state" with a message "m2" receive by an agent "name2" in a state "state2". If an unification exist, then an unification is save in match_message_story
@param name of the sender
@param state of the sender
@param m1 is the message sent by the sender
@param name2 is the name of the receiver
@param state2 is the state of the receiver
@param m2 is the message received by the receiver
@return bool
*)
 let match_message (name : string) (state : string) (m1 : string) (name2 : string) (state2 : string) (m2 : string) : bool =
 	let rec adjust_pair m =
		let rec adjust m =
			let decomp = decompose m in
			if(decomp=[]) then m
			else if((car decomp)="pair") then (list2string "," (List.map adjust (cdr decomp)))
			else ((car decomp)^"("^(list2string "," (List.map adjust_pair (cdr decomp)))^")")
		in	
		let decomp = decompose m in
		if(decomp=[])
		then m
		else(
			if((car decomp)="pair")
			then((car decomp)^"("^(list2string "," (List.map adjust  (cdr decomp)))^")")
			else ((car decomp)^"("^(list2string "," (List.map adjust_pair (cdr decomp)))^")")
		)
	in
	let rec zip_pair m =
		let rec aux l =
			match l with [] -> ""
			|[x] -> (zip_pair x)
			|(x::ls) -> "pair("^(zip_pair x)^","^(aux ls)^")"
		in
		let decomp = decompose m in
		if(decomp=[]) then m
		else if((car decomp)="pair") then (aux (cdr decomp))
		else (zip_pair (car decomp))^"("^(list2string "," (List.map zip_pair (cdr decomp)))^")"

	in
	let (exist_match,sub) = get_match_message_story name state m1 name2 state2 m2 (!untyped_variable) in
	if(exist_match<>(-1))
	then(
			if (sub=[]) then false
			else true
		)
	else(   	
		let t1 = get_type_of_var_in_state name state m1 in
		let t2 = get_type_of_var_in_state name2 state2 m2 in
		if(not(compatible t1 t2))
		then (add_in_match_message_story name state m1 name2 state2 m2 (!untyped_variable) []; false)
		else(
			let m01 = adjust_pair m1 in
			let m02 = adjust_pair m2 in
			let l_sub = unify name m01 name2 m02 !untyped_variable in
			let l_sub_res = (List.map 
					(fun (name1,value,name2,var) -> (name1,(zip_pair value),name2,var)) 
					l_sub)
			in
			add_in_match_message_story name state m1 name2 state2 m2 (!untyped_variable) l_sub_res;
			(l_sub<>[])
		)
	)		
;;


(** unify a state "m1" of an agent "name" with a state "m2" of an agent "name2". If an unification exist, then an unification is save in match_message_story. In all case, this function is used with name=name2.
@param name of the of the first agent
@param m1 is the state of the first agant
@param name2 is the name of the second agent
@param m2 is the state of the second agent
@return bool
*)
 let match_state (name : string) (m1 : string) (name2 : string) (m2 : string) : bool =
	let rec zip_pair m =
		let rec aux l =
			match l with [] -> ""
			|[x] -> (zip_pair x)
			|(x::ls) -> "pair("^(zip_pair x)^","^(aux ls)^")"
		in
		let decomp = decompose m in
		if(decomp=[]) then m
		else if((car decomp)="pair") then (aux (cdr decomp))
		else (zip_pair (car decomp))^"("^(list2string "," (List.map zip_pair (cdr decomp)))^")"

	in
	let (exist_match,sub) = get_match_message_story name m1 m1 name2 m2 m2 (!untyped_variable) in
	if(exist_match<>(-1))
	then ( 
		if(sub=[]) then false
		else true
	)
	else (
		let trans m = "pair("^m^")"
		in
		let d1 = trans m1 in 
		let d2 = trans m2 in
(*		let res = match_message name m1 d1 name2 m2 d2 in
		let (exist_match,sub) = get_match_message_story name m1 d1 name2 m2 d2 (!untyped_variable) in
		if (exist_match<>(-1))
		then add_in_match_message_story name m1 m1 name2 m2 m2 (!untyped_variable) sub;
		res*)

		let t1 = get_type_of_var_in_state name m1 d1 in
		let t2 = get_type_of_var_in_state name2 m2 d2 in
		if(not(compatible_state t1 t2))
		then (add_in_match_message_story name m1 d1 name2 m2 d2 (!untyped_variable) []; false)
		else(
			let l_sub = unify name d1 name2 d2 !untyped_variable in
			let l_sub_res = (List.map 
					(fun (name1,value,name2,var) -> (name1,(zip_pair value),name2,var)) 
					l_sub)
			in
			add_in_match_message_story name m1 m1 name2 m2 m2 (!untyped_variable) l_sub_res;
			(l_sub<>[])
		)

	)
 ;;


(** Assign values of variables in Message and in States 
@param name_to_assign is the name of the role which the value of variables are assigned
@param sub is a substitution. Each element is a 4-uplet of string where (name1, value, name2, variable) "value" is a value or a variable of the role "name1", and this value is assigned to the variable "variable" of the role "name2"
@return  unit
*)
  let assign_sub (name_to_assign : string) (sub : (string * string * string * string) list) : unit =
	let rec normalize name v = 
		let decomp = decompose v in
		if(decomp<>[]) then ((normalize name (car decomp))^"("^(list2string "," (List.map (fun e -> normalize name e) (cdr decomp)))^")")
		else if (is_a_var v) then normalize name (get_value_of_var name v)
		else v
	in
	List.iter
	(fun (name1,value,name2,var) ->
		if(name2=name_to_assign)
			then	  add_var_value name1 (normalize name1 value) name2 var
	)
	sub
 ;;
 (** Return an unification of the values of the variables in message "m2" of  an agent "name2" in state "state2" from the variables and the value of a message "m1" of an agent "name" in state "state"
@param name is the name of an agent
@param state is a state of the agent "name"
@param m1 is the message send by the agent "name"
@param name2 is the name of an agent
@param state2 is a state of the agent "name2"
@param m2 is a message receive by the agent "name2"
@return unit
*)
 let get_unification (name : string) (state : string) (m1 : string) (name2 : string) (state2 : string) (m2 : string) : (int * (string * string * string * string) list) =
  	let (exist_match,sub) = get_match_message_story name state m1 name2 state2 m2 (!untyped_variable) in
	if(exist_match<>(-1))
	then (exist_match,sub)
	else ( 
		let res = match_message name state m1 name2 state2 m2 in
		if(res)
		then( 
			get_match_message_story name state m1 name2 state2 m2 (!untyped_variable)
		)
		else (0,[])
	)	
 ;;
(** Assign the values of the variables in message "m2" of  an agent "name2" in state "state2" from the variables and the value of a message "m1" of an agent "name" in state "state"
@param name is the name of an agent
@param state is a state of the agent "name"
@param m1 is the message send by the agent "name"
@param name2 is the name of an agent
@param state2 is a state of the agent "name2"
@param m2 is a message receive by the agent "name2"
@return unit
*)
 let assign_var (name : string) (state : string) (m1 : string) (name2 : string) (state2 : string) (m2 : string) : unit =
  	let (exist_match,sub) = get_match_message_story name state m1 name2 state2 m2 (!untyped_variable) in
	if(exist_match<>(-1))
	then (
		if(sub<>[]) then assign_sub name2 sub
	)
	else (
		let res = match_message name state m1 name2 state2 m2 in
		if(res)then(
			let (exist_match,sub) = get_match_message_story name state m1 name2 state2 m2 (!untyped_variable) in
			if(exist_match<>(-1)) then assign_sub name2 sub
		)
	) 
 ;;
(** Assign the values of the variables of an agent "name2" in state "state2" from the state "state" of an agent "name".  In all case, this function is used with name=name2.
@param name is the name of an agent
@param state is a state of the agent "name"
@param name2 is the name of an agent
@param state2 is a state of the agent "name2"
@return unit
*)
 let assign_var_of_state (name : string) (m1 : string) (name2 : string) (m2 : string) : unit = 
	let (exist_match,sub) = get_match_message_story name m1 m1 name2 m2 m2 (!untyped_variable) in
	if(exist_match<>(-1))
	then (
		if(sub<>[]) then assign_sub name2 sub
	)
	else (
		let rec trans l =	
			match l with [] -> ""
				|[x] -> x
				|(x::ls)	-> "pair("^x^","^(trans ls)^")"
		in
		let d1 = trans (cdr(decompose ("x("^m1^")"))) in
		let d2 = trans (cdr(decompose ("x("^m2^")"))) in
		let res = match_message name m1 d1 name2 m2 d2 in
		if(res)
		then(
			let (exist_match,sub) = get_match_message_story name m1 d1 name2 m2 d2 (!untyped_variable) in
			if(exist_match<>(-1))
			then( 	
				add_in_match_message_story name m1 m1  name2 m2 m2 (!untyped_variable) sub;
				assign_sub name2 sub
			)
		)
	) 
 ;;

(**************************************************************************)

(** 
test if the value or the variable "value" of the role "name0" match with an elemement of the list "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type list
@return a boolean
*)
 let in_list (name0 : string) (value : string) (name : string) (var : string) : bool =
	let rec member_match x l =
		match l with [] -> false
			|(e::ls) -> 	if (match_message name0 x x name e e ) 
					then true 
					else (member_match x ls);
	in
 	if(is_a_list var)
	then(
		let list_to_refer = if(is_a_var var) then get_value_of_var name var else var in
		let old_value_of_list = if(Hashtbl.mem var_value ("x",list_to_refer)) then (Hashtbl.find var_value ("x",list_to_refer)) else "" in
		let val_to_add = value
			(*if((is_a_var value))
			then (get_value_of_var name0 value)
 			else value *)
		in
		let decomp = decompose old_value_of_list in
		(decomp<>[]) && (member_match val_to_add (cdr decomp)) (* (List.mem val_to_add (cdr decomp)) *)
	)
	else false
;;

(** 
test if the value or the variable "value" of the role "name0" is equal to an elemement of the list "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type list
@return a boolean
*)
 let in_list_eq (name0 : string) (value : string) (name : string) (var : string) : bool =
	let rec member_match x l =
		match l with [] -> false
			|(e::ls) -> 	if (e=x) 
					then true 
					else (member_match x ls);
	in
	let rec decomp_exp exp0 =
		let exp = if (is_a_var exp0) then (get_value_of_var name0 exp0) else exp0 in
		let d = decompose exp in
		if(d=[] && ((List.length d)<1)) then exp
		else (car d)^"("^(list2string "," (List.map (fun e -> decomp_exp e) (cdr d)))^")"
	in
 	if(is_a_list var)
	then(
		let list_to_refer = if(is_a_var var) then get_value_of_var name var else var in
		let old_value_of_list = if(Hashtbl.mem var_value ("x",list_to_refer)) then (Hashtbl.find var_value ("x",list_to_refer)) else "" in
		let val_to_add = decomp_exp value
		in
		let decomp = decompose old_value_of_list in
		(decomp<>[]) && (member_match val_to_add (cdr decomp)) (* (List.mem val_to_add (cdr decomp)) *)
	)
	else false
;;

(** 
assign the values of the variables which composes a message if the values of the variable "value" of the role "name0" match with an elemement of the list "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type list
@return a boolean
*)
 let assign_in_list (name0 : string) (value : string) (name : string) (var : string) : bool =
	let rec member_match x l =
		match l with [] -> false
			|(e::ls) -> if (match_message name e e name0 x x ) then ((assign_var name e e name0 x x );true) else (member_match x ls);
	in
 	if(is_a_list var)
	then(
		let list_to_refer = if(is_a_var var) then get_value_of_var name var else var in
		let old_value_of_list = if(Hashtbl.mem var_value ("x",list_to_refer)) then (Hashtbl.find var_value ("x",list_to_refer)) else "" in
		let val_to_add = value
			(*if((is_a_var value))
			then (get_value_of_var name0 value)
 			else value *)
		in
		let decomp = decompose old_value_of_list in
		(decomp<>[]) && (member_match val_to_add (cdr decomp)) (* (List.mem val_to_add (cdr decomp)) *)
	)
	else false
;;

(** 
test if the value or tha variable "value" of the role "name0" match with an element in the set "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type set
@return a boolean
*)
 let in_set (name0 : string) (value : string) (name : string) (var : string) : bool =
	let rec member_match x l =
		match l with [] -> false
			|(e::ls) -> 	if (match_message  name e e name0 x x) 
					then true 
					else (member_match x ls);
	in
 	if(is_a_set var)
	then(
		let set_to_refer = if(is_a_var var) then get_value_of_var name var else var in
		let old_value_of_set = if(Hashtbl.mem var_value ("x",set_to_refer)) then (Hashtbl.find var_value ("x",set_to_refer)) else "" in
		let val_to_add = value
			(*if((is_a_var value))
			then (get_value_of_var name0 value)
 			else value *)
		in
		let decomp = decompose old_value_of_set in
		(decomp<>[]) && (member_match val_to_add (cdr decomp)) (* && (List.mem val_to_add (cdr decomp)) *)
	)
	else false
;;

(** 
test if the value or tha variable "value" of the role "name0" is equal to an element in the set "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type set
@return a boolean
*)
 let in_set_eq (name0 : string) (value : string) (name : string) (var : string) : bool =
	let rec member_match x l =
		match l with [] -> false
			|(e::ls) -> 	if (e=x)
					then true 
					else (member_match x ls);
	in
	let rec decomp_exp exp0 =
		let exp = if (is_a_var exp0) then (get_value_of_var name0 exp0) else exp0 in
		let d = decompose exp in
		if(d=[] && ((List.length d)<1)) then exp
		else (car d)^"("^(list2string "," (List.map (fun e -> decomp_exp e) (cdr d)))^")"
	in
 	if(is_a_set var)
	then(
		let set_to_refer = if(is_a_var var) then get_value_of_var name var else var in
		let old_value_of_set = if(Hashtbl.mem var_value ("x",set_to_refer)) then (Hashtbl.find var_value ("x",set_to_refer)) else "" in
		let val_to_add = decomp_exp value 
		in
		let decomp = decompose old_value_of_set in
		(decomp<>[]) && (member_match val_to_add (cdr decomp)) (* && (List.mem val_to_add (cdr decomp)) *)
	)
	else false
;;

(** 
assign the values of the variables which composes a message if the values of the variable "value" of the role "name0" match with an elemement of the set "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type set
@return a boolean
*)
 let assign_in_set (name0 : string) (value : string) (name : string) (var : string) : bool =
	let rec member_match x l =
		match l with [] -> false
			|(e::ls) -> if (match_message name e e name0 x x) then ((assign_var name e e name0 x x );true) else (member_match x ls);
	in
 	if(is_a_set var)
	then(
		let set_to_refer = if(is_a_var var) then get_value_of_var name var else var in
		let old_value_of_set = if(Hashtbl.mem var_value ("x",set_to_refer)) then (Hashtbl.find var_value ("x",set_to_refer)) else "" in
		let val_to_add = value
			(*if((is_a_var value))
			then (get_value_of_var name0 value)
 			else value *)
		in
		let decomp = decompose old_value_of_set in
		(decomp<>[]) && (member_match val_to_add (cdr decomp)) (* && (List.mem val_to_add (cdr decomp)) *)
	)
	else false
;;

 (** 
assign the values of the variables which composes a message if the values of the variable "value" of the role "name0" match with an elemement of the aggregation entity "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type set or list
@return a boolean
*)
 let in_aggregation (name0 : string) (value : string) (name : string) (var : string) : bool =
 	(in_set name0 value name var) || (in_list name0 value name var)
 ;;

 (** 
assign the values of the variables which composes a message if the values of the variable "value" of the role "name0" is equal to an elemement of the aggregation entity "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type set or list
@return a boolean
*)
 let in_aggregation_eq (name0 : string) (value : string) (name : string) (var : string) : bool =
 	(in_set_eq name0 value name var) || (in_list_eq name0 value name var)
 ;;

 (** 
test if the value or the variable "value" of the role "name0" is contained in "var" of the role "name"
@param name0 is an agent
@param value is a value or a variable
@param name is an agent
@param var is a variable of type set or list
@return a boolean
*)
 let assign_in_aggregation (name0 : string) (value : string) (name : string) (var : string) : bool =
 	(assign_in_set name0 value name var) || (assign_in_list name0 value name var)
 ;;

(** treatment of sets and lists and other operations (not, equal, /=, cons, delete, in) *)

(** this function finds the operation associated with a state describe in a right member of a transition
@param name if the name of an agent
@param state is a state describe in a right member of a transition
@return (op_left, op_right) where op_left is the operations of left member of transition, and op_right is the operations of the right member of the transition
*)
let find_operation_of_state (name : string) (state : string) : (string list * string list) * (string list * string list) =
	let rec aux lActions =
		match lActions with [] -> (([],[]),([],[]))
			|((_,p,final)::ls) -> if(state=(findState final)) then p else aux ls;	
	in
	let rec find_operation l_automate  =
		match l_automate with [] -> (([],[]),([],[]))
			|((name2,id2,_,lActions)::ls) -> 	if(name=(get_name_in_session name2 id2))
								then aux lActions
								else find_operation ls;
	in find_operation (!automate)
;;
(** this function return a boolean value. If this value is true then the tests operations are satisfied, else if this value is false then the tests operations are not satisfied
@param name if the name of an agent
@param state is a state describe in a right member of a transition
@return bool
*)
let operations_test_treatment_left (name : string) (state : string) : bool =
	let rec test t =
		let rec var2value_in_exp exp =
			let decomp = decompose exp in
			if(decomp<>[]) then ((car decomp)^"("^(list2string "," (List.map var2value_in_exp (cdr decomp)))^")")
			else if(is_a_value exp) then exp
			else if(is_a_var exp) then (get_value_of_var name exp)
			else exp
		in

		let rec test0 t =
	
			let rec test_aux exp =
				let decomp = decompose exp in
				match decomp with [] -> if(is_a_var exp) then (get_value_of_var name exp) else exp
						|x::ls -> if(List.mem x ["not" ; "contains" ; "idequal" ; "equal"])
							then (string_of_bool (test (x::ls)))
							else (x^"("^(list2string "," (List.map test_aux ls))^")")
			in
			match t with [] -> true
				|(x::ls) -> 	let d = decompose x in
						if(d<>[])
						then(
							if((car d)="contains") then(
								let arg1 = if(is_a_var (cadr d)) then (get_value_of_var name (cadr d)) else (cadr d) in
								let arg2 = if(is_a_var (caddr d)) then (get_value_of_var name (caddr d)) else (caddr d) in
								(in_aggregation name arg1 name arg2) || (in_aggregation name (var2value_in_exp arg1) name arg2)
							)
							else if((car d)="equal" || (car d)="idequal") then(
								let arg1 = 	if(is_a_var (cadr d)) then (get_value_of_var name (cadr d)) 
										else if(is_a_value (cadr d)) then (cadr d) 
										else (test_aux (cadr d)) in
								let arg2 =  	if(is_a_var (caddr d)) then (get_value_of_var name (caddr d)) 
										else if(is_a_value (caddr d)) then (caddr d) 
										else (test_aux (caddr d) )in
								((match_message name state arg1 name state arg2) || (match_message name state arg2 name state arg1))  && (test ls)
							)
							else if((car d)="leq") then (
								let arg1 = if(is_a_var (cadr d)) then (get_value_of_var name (cadr d)) else (cadr d) in
								let arg2 = if(is_a_var (caddr d)) then (get_value_of_var name (caddr d)) else (cadr d) in
								((int_of_string arg1)<=(int_of_string arg2)) && (test ls)
							)
							else if((car d)="not") then (
								not(test [(cadr d)]) && (test ls)
							)
							else (test0 ls);
						)
						else (test0 ls);
		in

		let rec test_aux exp =
			let decomp = decompose exp in
			match decomp with [] -> exp
					|x::ls -> if(List.mem x ["not" ; "contains" ; "idequal" ; "equal"])
						then (string_of_bool (test (x::ls)))
						else (x^"("^(list2string "," (List.map test_aux ls))^")")
		in
		match t with [] -> true
			|(x::ls) -> 	let d = decompose x in
					if(d<>[])
					then(
						if((car d)="contains") then(
							let arg1 = if(is_a_var (cadr d)) then (get_value_of_var name (cadr d)) else (cadr d) in
							let arg2 = if(is_a_var (caddr d)) then (get_value_of_var name (caddr d)) else (caddr d) in
							(in_aggregation name arg1 name arg2) || (in_aggregation name (var2value_in_exp arg1) name arg2)
						)
						else if((car d)="equal" || (car d)="idequal") then(
							let arg1 = 	if(is_a_var (cadr d)) then (get_value_of_var name (cadr d)) 
									else if(is_a_value (cadr d)) then (cadr d) 
									else (test_aux (cadr d)) in
							let arg2 =  	if(is_a_var (caddr d)) then (get_value_of_var name (caddr d)) 
									else if(is_a_value (caddr d)) then (caddr d) 
									else (test_aux (caddr d) )in
							((match_message name state arg1 name state arg2) || (match_message name state arg2 name state arg1))  && (test ls)
						)
						else if((car d)="leq") then (
							let arg1 = if(is_a_var (cadr d)) then (get_value_of_var name (cadr d)) else (cadr d) in
							let arg2 = if(is_a_var (caddr d)) then (get_value_of_var name (caddr d)) else (cadr d) in
							((int_of_string arg1)<=(int_of_string arg2)) && (test ls)
						)
						else if((car d)="not") then (
							not(test0 [(cadr d)]) && (test ls)
						)
						else (test ls);
					)
					else (test ls);
	in
	match (find_operation_of_state name state) with ((gt,_),_) -> (test gt)
;;
(** this function treats this affections operations like add an element in a list or in a set, and delete an element in a list or in a set
@param name if the name of an agent
@param state is a state describe in a right member of a transition
@return bool
*)
let operations_test_treatment_right (name : string) (state : string) : bool =
	let rec test t =
		let rec test_aux exp =
			let decomp = decompose exp in
			match decomp with [] -> exp
					|x::ls -> if(List.mem x ["not" ; "contains" ; "idequal" ; "equal"])
						then (string_of_bool (test (x::ls)))
						else (x^"("^(list2string "," (List.map test_aux ls))^")")
		in
		let rec var2test exp =
			let decomp = decompose exp in
			if(decomp<>[]) then ((car decomp)^"("^(list2string "," (List.map var2test (cdr decomp)))^")")
			else if(is_a_value exp) then exp
			else if(is_a_test_var exp) then exp
			else if(is_a_var exp) then ("Test_"^exp)
			else exp
		in
		match t with [] -> true
			|(x::ls) -> 	let d = decompose x in
					if(d<>[])
					then(
						if((car d)="contains") then(
							in_aggregation name (cadr d) name (caddr d)
						)
						else if((car d)="equal" || (car d)="idequal") then(
							let arg1 = 	if(is_a_test_var (cadr d)) then (get_value_of_var name (cadr d)) 
									else if(is_a_value (cadr d)) then (cadr d) 
									else (string_of_bool (test [(cadr d)] ))in
							let arg2 =  	if(is_a_test_var (caddr d)) then (get_value_of_var name (caddr d)) 
									else if(is_a_value (caddr d)) then (caddr d) 
									else (test_aux (caddr d) )in
							((match_message name state (var2test arg1) name state (var2test arg2)) || (match_message name state (var2test arg2) name state (var2test arg1))) && (test ls)
						)
						else if((car d)="leq") then (
							let arg1 = if(is_a_var (cadr d)) then (get_value_of_var name (cadr d)) else (cadr d) in
							let arg2 = if(is_a_var (caddr d)) then (get_value_of_var name (caddr d)) else (caddr d) in
							((int_of_string arg1)<=(int_of_string arg2)) && (test ls)
						)
						else if((car d)="not") then (
							not(test [(cadr d)]) && (test ls)
						)
						else (test ls)
					)
					else (test ls);
	in
	match (find_operation_of_state name state) with (_,(dt,_)) -> (test dt)
;;
(** this function return a boolean value. If this value is true then the tests operations are satisfied, else if this value is false then the tests operations are not satisfied
@param name if the name of an agent
@param state is a state describe in a right member of a transition
@return unit
*)
let operations_affectation_treatment (name : string) (state : string) : unit =


	let assign_var (name : string) (state : string) (m1 : string) (name2 : string) (state2 : string) (m2 : string) : (int * (string * string * string * string) list) =
		let (exist_match,sub) = get_match_message_story name state m1 name2 state2 m2 (!untyped_variable) in
		if(exist_match<>(-1))
		then (
		(exist_match,sub)	
		)
		else (
			let res = match_message name state m1 name2 state2 m2 in
			if(res)then(
				let (exist_match,sub) = get_match_message_story name state m1 name2 state2 m2 (!untyped_variable) in
				 (exist_match,sub)
			)
			else (-1,[])
		) 
	in

	let rec affect a =
		match a with [] -> ()
			|(x::ls) -> 	let d = decompose x in
					if(d<>[] && ((List.length d)>2))
					then(	
						let arg1 = (cadr d) in
						let arg2 = (caddr d) in
						if((is_a_var arg2) && (car d)="cons") then(
							add_in_aggregation name arg1 name arg2
						)
						else if((is_a_var arg2) && (car d)="delete") then(
							delete_in_aggregation name arg1 name arg2
						)
						else if((car d)="contains") then(
							ignore(assign_in_aggregation name (cadr d) name (caddr d))
						)
						else if((car d)="equal" || (car d)="idequal") then(
							let arg1 = 	if(is_a_test_var (cadr d)) then (get_value_of_var name (cadr d)) 
									else if(is_a_value (cadr d)) then (cadr d) 
									else (cadr d) in
							let arg2 =  	if(is_a_test_var (caddr d)) then (get_value_of_var name (caddr d)) 
									else if(is_a_value (caddr d)) then (caddr d) 
									else (caddr d) in
							let (exist_match1,sub1) = assign_var name state arg1 name state arg2 in
							let (exist_match2,sub2) = assign_var name state arg2 name state arg1 in
							if(exist_match1<>(-1) && sub1<>[]) then assign_sub name sub1;
							if(exist_match2<>(-1) && sub2<>[]) then assign_sub name sub2;
						);
					)
					else 	if(d<>[] && ((List.length d)>1) && (car d)="not")
						then(	let dd = decompose (cadr d) in
							if(dd<>[] && (car dd)="not") then (
							ignore(affect [(cadr dd)])
							);
						);
					affect ls;
	in
	match (find_operation_of_state name state) with ((la,ga),(_,da)) -> ignore((affect la) ; (affect ga) ; (affect da))
;;
